# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import tkinter as tk

import octadist


class ScriptingConsole:
    """
    Start scripting interface for an interactive code.

    User can access to class variable (dynamics variable).

    +------------+
    | Output box |
    +------------+
    | Input box  |
    +------------+

    Parameters
    ----------
    root : object
        Passing self object from another class to this class as root argument.

    See Also
    --------
    settings :
        Program settings.

    Examples
    --------
    >>> import tkinter as tk
    >>> master = tk.Tk()
    >>> console = ScriptingConsole(master)
    >>> console.scripting_start()

    """
    def __init__(self, root):
        self.root = root
        self.wd = None
        self.history_command = []

    def scripting_start(self):
        """
        Start scripting console.

        """

        self.wd = tk.Toplevel(self.root.master)
        if self.root.octadist_icon is not None:
            self.wd.wm_iconbitmap(self.root.octadist_icon)
        self.wd.title("OctaDist Scripting Interface")
        self.wd.bind('<Return>', self.script_execute)
        self.wd.resizable(0, 0)

        lbl = tk.Label(self.wd, text="Output:")
        lbl.grid(padx="5", pady="5", sticky=tk.W, row=0, column=0)
        self.box_script = tk.Text(self.wd, width=70, height=20)
        self.box_script.grid(padx="5", pady="5", row=1, column=0, columnspan=2)
        lbl = tk.Label(self.wd, text="Input:")
        lbl.grid(padx="5", pady="5", sticky=tk.W, row=2, column=0)
        self.entry_script = tk.Entry(self.wd, width=62)
        self.entry_script.grid(padx="5", pady="5", sticky=tk.W, row=3, column=0)
        btn_script = tk.Button(self.wd, text="Run")
        btn_script.bind('<Button-1>', self.script_execute)
        btn_script.grid(padx="5", pady="5", row=3, column=1)

        self.box_script.insert(tk.INSERT, "Welcome to OctaDist interactive scripting console\n")
        self.box_script.insert(tk.INSERT, "If you have no idea what to do about scripting, "
                                          "type \"help\" to get started.\n\n")

        self.wd.mainloop()

    def script_run_help(self):
        """
        Show help messages.

        """
        help_msg = f"""\
>>> Interactive code console for OctaDist.
>>> This scripting interface supports built-in commands as follows:
>>> 
>>> Command\t\tDescription
>>> =======\t\t===========
>>> help\t\tShow this help info.
>>> list\t\tList all commands.
>>> info\t\tShow info of program.
>>> doc\t\tShow docstring of this function.
>>> show\t\tShow values of variables.
>>> \t\tUsage: show var1 [var2] [var3] [...]
>>> type\t\tShow type of variables.
>>> \t\tUsage: type var1 [var2] [var3] [...]
>>> set\t\tSet new value to variable.
>>> \t\tUsage: set var value
>>> clear\t\tClear stdout/stderr.
>>> clean\t\tClear stdout/stderr and command history.
>>> \t\thistory - Clean previous commands history.
>>> \t\tmaster - Clean cache of OctaDist master class.
>>> restore\t\tRestore program settings.
>>> history\t\tCommand history.
>>> 
>>> Rangsiman Ketkaew\t\t<rangsiman1993@gmail.com>\t\tOctaDist {octadist.__version__}
"""

        self.box_script.insert(tk.INSERT, help_msg + "\n")

    def script_run_list(self):
        """
        Show list of commands in scripting run.

        """
        all_command = "help, list, info, doc, show, type, set, " \
                      "clear, clean, restore, history, exit"
        self.box_script.insert(tk.INSERT, f">>> {all_command}\n")

    def script_run_info(self):
        """
        Show info of program.

        """
        self.box_script.insert(tk.INSERT, f">>> {octadist.__description__}\n")

    def script_run_doc(self):
        """
        Show document of program.

        """
        self.box_script.insert(tk.INSERT, f">>> {octadist.__doc__}\n")

    def script_run_show(self, args):
        """
        Show value of variable that user requests.

        Parameters
        ----------
        args : str
            Arbitrary argument.

        """
        if not args:
            self.box_script.insert(tk.INSERT, ">>> show command needs at least 1 parameter\n")
            return 1

        first_arg = args[0].lower()
        if first_arg == "all" or first_arg == "*":
            for key in self.root.__dict__.keys():
                value = self.root.__dict__[key]
                self.box_script.insert(tk.INSERT, f">>> {key} = {value}\n")
            return 0

        for i in range(len(args)):
            try:
                key = args[i]
                value = self.root.__dict__[f"{key}"]
                self.box_script.insert(tk.INSERT, f">>> {key} = {value}\n")
            except KeyError:
                self.box_script.insert(tk.INSERT, f">>> variable \"{args[i]}\" is not defined\n")

    def script_run_type(self, args):
        """
        Show data type of variable.

        Parameters
        ----------
        args : str
            Arbitrary argument.

        """
        if not args:
            self.box_script.insert(tk.INSERT, ">>> \"type\" command needs at least 1 parameter\n")
            return 1

        for i in range(len(args)):
            try:
                key = args[i]
                data_type = type(self.root.__dict__[f"{key}"])
                self.box_script.insert(tk.INSERT, f">>> {key} : {data_type}\n")
            except KeyError:
                self.box_script.insert(tk.INSERT, f">>> variable \"{args[i]}\" is not defined\n")

    def script_run_set(self, args):
        """
        Set new value to variable.

        Parameters
        ----------
        args : str
            Arbitrary argument.

        """
        if len(args) != 2:
            self.box_script.insert(tk.INSERT, ">>> No variable specified\n")
            self.box_script.insert(tk.INSERT, ">>> set command needs 2 parameters\n")
            return 1

        key = args[0]
        value = args[1]

        self.root.__dict__[f"{key}"] = value

        if f"key" not in self.root.__dict__:
            self.box_script.insert(tk.INSERT, f">>> new dynamic variable \"{key}\" defined\n")

        self.box_script.insert(tk.INSERT, f">>> {key} = {value}\n")

    def script_run_clear(self):
        """
        Clear output box.

        """
        self.box_script.delete(1.0, tk.END)

    def script_run_clean(self, args):
        """
        Clear output box and clean variable.

        """
        if not args:
            # clean scripting console
            self.script_run_clear()
        elif args[0].lower() == "history":
            self.history_command = []
        elif args[0].lower() == "master":
            # clean master windows
            self.root.clear_cache()

    def script_run_restore(self):
        """
        Restore all default settings.

        """
        self.root.cutoff_metal_ligand = self.root.backup_cutoff_metal_ligand
        self.root.cutoff_global = self.root.backup_cutoff_global
        self.root.cutoff_hydrogen = self.root.backup_cutoff_hydrogen
        self.root.text_editor = self.root.backup_text_editor
        self.root.show_title = self.root.backup_show_title
        self.root.show_axis = self.root.backup_show_axis
        self.root.show_grid = self.root.backup_show_grid

        self.box_script.insert(tk.INSERT, f">>> Restore all settings\n")

        var_settings = ("cutoff_metal_ligand",
                        "cutoff_global",
                        "cutoff_hydrogen",
                        "text_editor",
                        "show_title",
                        "show_axis",
                        "show_grid"
                        )

        for key in var_settings:
            value = self.root.__dict__[f"{key}"]
            self.box_script.insert(tk.INSERT, f">>> {key} = {value}\n")

    def script_run_history(self):
        """
        Show history of command.

        """
        if not self.history_command:
            self.box_script.insert(tk.INSERT, f">>> no history\n")
        for item in self.history_command:
            self.box_script.insert(tk.INSERT, f">>> {item}\n")

    def script_no_command(self, command):
        """
        Show statement if command not found.

        Parameters
        ----------
        command : str
            Command that user submits.

        """
        self.box_script.insert(tk.INSERT, f">>> Command \"{command}\" not found\n")

    def script_execute(self, event):
        """
        Execute input command scripting.

        Parameters
        ----------
        event : object
            Object for button interaction

        """
        user_command = self.entry_script.get().strip().split()
        self.entry_script.delete(0, tk.END)

        # print input command
        if user_command:
            res = (" ".join([str(i) for i in user_command]))
            self.box_script.insert(tk.INSERT, f"{res}\n")
            # collect command to history list
            if res.lower() != "history":
                self.history_command.append(res)

        if len(user_command) == 0:
            self.box_script.insert(tk.INSERT, ">>> no command\n")
            self.box_script.see(tk.END)
            return 1

        command = user_command[0].lower()
        args = user_command[1:]

        if command == "help":
            self.script_run_help()
        elif command == "list":
            self.script_run_list()
        elif command == "info":
            self.script_run_info()
        elif command == "doc":
            self.script_run_doc()
        elif command == "show":
            self.script_run_show(args)
        elif command == "type":
            self.script_run_type(args)
        elif command == "set":
            self.script_run_set(args)
        elif command == "clear":
            self.script_run_clear()
        elif command == "clean":
            self.script_run_clean(args)
        elif command == "restore":
            self.script_run_restore()
        elif command == "history":
            self.script_run_history()
        elif command == "exit" or command == "quit":
            self.wd.destroy()
            return 1
        else:
            self.script_no_command(command)

        self.box_script.see(tk.END)




# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import base64
import os
import platform
import subprocess
import tkinter as tk
import tkinter.scrolledtext as tkscrolled
import webbrowser
from tkinter import filedialog
from tkinter import messagebox
from tkinter import ttk
from tkinter.messagebox import showinfo
from urllib.request import urlopen

import numpy as np

import octadist
from octadist.logo import Icon_Base64
from octadist.src import (
    calc, molecule, draw, plot, popup, scripting, structure, tools
)


class OctaDist:
    """
    OctaDist class initiates main program UI and create all widgets.

    Program interface is structured as follows:

    +-------------------+
    |   Program Menu    |
    +-------------------+
    |     Frame 1       |
    +---------+---------+
    | Frame 2 | Frame 3 |
    +---------+---------+
    |     Frame 4       |
    +-------------------+

    - Frame 1 : Program name and short description
    - Frame 2 : Program console
    - Frame 3 : Textbox for showing summary output
    - Frame 4 : textbox for showing detailed output

    Examples
    --------
    >>> my_app = OctaDist()
    >>> my_app.start_app()

    """

    def __init__(self):
        self.master = tk.Tk()

        # Initialize parameters
        self.file_list = []  # Full path of input files.
        self.file_name = []  # File name.
        self.octa_index = []  # Octahedral structure index.
        self.atom_coord_full = []  # Coordinates of metal complex.
        self.atom_coord_octa = []  # Coordinates of octahedral structures.
        self.all_zeta = []  # Zeta of all octahedral structures.
        self.all_delta = []  # Delta of all octahedral structures.
        self.all_sigma = []  # Sigma of all octahedral structures.
        self.all_theta = []  # Theta of all octahedral structures.
        self.comp_result = []  # Distortion parameters.

        self.octadist_icon = None

        # Default cutoff values
        self.cutoff_metal_ligand = 2.8
        self.cutoff_global = 2.0
        self.cutoff_hydrogen = 1.2

        # Default executable of text editor
        self.text_editor = "notepad.exe"

        # Default display settings
        self.show_title = True
        self.show_axis = True
        self.show_grid = True

        self.backup_cutoff_metal_ligand = self.cutoff_metal_ligand
        self.backup_cutoff_global = self.cutoff_global
        self.backup_cutoff_hydrogen = self.cutoff_hydrogen
        self.backup_text_editor = self.text_editor
        self.backup_show_title = self.show_title
        self.backup_show_axis = self.show_axis
        self.backup_show_grid = self.show_grid

        # Create master frame, sub-frames, add menu, and add widgets
        self.create_logo()
        self.start_master()
        self.add_menu()
        self.add_widgets()
        self.welcome_msg()
        self.start_app()

    def create_logo(self):
        """
        Create icon file from Base64 raw code.

        This will be used only for Windows OS.

        Other OS like Linux and macOS use default logo of Tkinter.

        Examples
        --------
        >>> if self.octadist_icon is True:
        >>>     self.create_logo()
        >>> else:
        >>>     pass

        """
        if platform.system() == "Windows":
            icon_data = base64.b64decode(Icon_Base64.icon_base64)
            temp_file = "icon.ico"
            save_path = os.path.expanduser("~/AppData/Local/Temp")
            self.octadist_icon = os.path.join(save_path, temp_file)
            icon_file = open(self.octadist_icon, "wb")
            icon_file.write(icon_data)
            icon_file.close()
            self.master.wm_iconbitmap(self.octadist_icon)

    def start_master(self):
        """
        Start application with UI settings.

        """
        self.master.title(f"OctaDist {octadist.__version__}")
        font = "Arial 10"
        self.master.option_add("*Font", font)
        center_width = (self.master.winfo_screenwidth() / 2.) - (550 / 2.)
        center_height = (self.master.winfo_screenheight() / 2.) - (750 / 2.)
        self.master.geometry("525x635+%d+%d" % (center_width, center_height))
        self.master.resizable(0, 0)

    def add_menu(self):
        """
        Add menu bar to master windows.

        """
        # Main menu
        menu_bar = tk.Menu(self.master)
        self.master.config(menu=menu_bar)
        file_menu = tk.Menu(self.master, tearoff=0)
        edit_menu = tk.Menu(self.master, tearoff=0)
        disp_menu = tk.Menu(self.master, tearoff=0)
        tools_menu = tk.Menu(self.master, tearoff=0)
        help_menu = tk.Menu(self.master, tearoff=0)

        # Sub-menu
        copy_menu = tk.Menu(self.master, tearoff=0)
        data_menu = tk.Menu(self.master, tearoff=0)
        strct_menu = tk.Menu(self.master, tearoff=0)

        # File
        menu_bar.add_cascade(label="File", menu=file_menu)
        file_menu.add_command(label="New", command=lambda: self.clear_cache())
        file_menu.add_command(label="Open...", command=lambda: self.open_file())
        file_menu.add_command(label="Save Results", command=lambda: self.save_results())
        file_menu.add_command(label="Save Coordinates", command=lambda: self.save_coord())
        file_menu.add_separator()
        file_menu.add_command(label="Settings", command=lambda: self.settings())
        file_menu.add_separator()
        file_menu.add_command(label="Exit", command=lambda: self.master.destroy())

        # Edit
        menu_bar.add_cascade(label="Edit", menu=edit_menu)
        edit_menu.add_cascade(label="Copy... to clipboard", menu=copy_menu)
        copy_menu.add_command(label="File Name", command=lambda: self.copy_name())
        copy_menu.add_command(label="File Path", command=lambda: self.copy_path())
        copy_menu.add_command(label="Computed Distortion Parameters", command=lambda: self.copy_results())
        copy_menu.add_command(label="Coordinates of Octahedral Structure", command=lambda: self.copy_octa())
        edit_menu.add_separator()
        edit_menu.add_command(label="Edit File", command=lambda: self.edit_file())
        edit_menu.add_command(label="Run Scripting Console", command=lambda: self.scripting_console())
        edit_menu.add_separator()
        edit_menu.add_command(label="Clear All Results", command=lambda: self.clear_cache())

        # Display
        menu_bar.add_cascade(label="Display", menu=disp_menu)
        disp_menu.add_command(label="Complex", command=lambda: self.draw_all_atom())
        disp_menu.add_command(label="Complex and Eight Faces", command=lambda: self.draw_all_atom_and_face())
        disp_menu.add_separator()
        disp_menu.add_command(label="Octahedron", command=lambda: self.draw_octa())
        disp_menu.add_command(label="Octahedron and Eight Faces", command=lambda: self.draw_octa_and_face())
        disp_menu.add_separator()
        disp_menu.add_command(label="Projection Planes", command=lambda: self.draw_projection())
        disp_menu.add_command(label="Twisting Triangular Faces", command=lambda: self.draw_twisting_plane())

        # Tools
        menu_bar.add_cascade(label="Tools", menu=tools_menu)
        tools_menu.add_cascade(label="Data Summary", command=lambda: self.show_data_complex())
        tools_menu.add_cascade(label="Show Structural Parameter", command=lambda: self.show_param_octa())
        tools_menu.add_command(label="Calculate Surface Area", command=lambda: self.show_surface_area())
        tools_menu.add_separator()
        tools_menu.add_command(label="Relationship Plot between ζ and Σ", command=lambda: self.plot_zeta_sigma())
        tools_menu.add_command(label="Relationship Plot between Σ and Θ", command=lambda: self.plot_sigma_theta())
        tools_menu.add_separator()
        tools_menu.add_command(label="Calculate Jahn-Teller Distortion", command=lambda: self.tool_jahn_teller())
        tools_menu.add_command(label="Calculate RMSD", command=lambda: self.tool_rmsd())

        # Help
        menu_bar.add_cascade(label="Help", menu=help_menu)
        help_menu.add_command(label="Quick Help", command=lambda: self.show_help())
        help_menu.add_command(label="User Documentation",
                              command=lambda: webbrowser.open_new_tab(octadist.__help__))
        help_menu.add_separator()
        submit_issue = "https://github.com/OctaDist/OctaDist/issues"
        help_menu.add_command(label="Report Issue", command=lambda: webbrowser.open_new_tab(submit_issue))
        help_menu.add_command(label="Github Repository",
                              command=lambda: webbrowser.open_new_tab(octadist.__github__))
        help_menu.add_command(label="Homepage", command=lambda: webbrowser.open_new_tab(octadist.__website__))
        help_menu.add_separator()
        help_menu.add_command(label="License", command=lambda: self.show_license())
        help_menu.add_separator()
        help_menu.add_command(label="Check for Updates...", command=self.check_update)
        help_menu.add_command(label="About Program", command=lambda: self.show_about())

    def add_widgets(self):
        """
        Add all widgets and components to master windows.

        GUI style of widgets in master windows use ttk style.

        """
        # my personal ttk style #
        gui_ttk = ttk.Style()
        gui_ttk.configure("TButton", relief="sunken", padding=5)
        gui_ttk.configure("My.TLabel", foreground="black")
        gui_ttk.configure("My.TLabelframe", foreground="brown")

        ###########
        # Frame 1 #
        ###########

        frame1 = tk.Frame(self.master)
        frame1.grid(padx=5, pady=5, row=0, column=0, columnspan=2)

        title = octadist.__title__
        lbl = tk.Label(frame1, text=title)
        lbl.configure(foreground="blue", font=("Arial", 16, "bold"))
        lbl.grid(pady="5", row=0, columnspan=4)

        description = octadist.__description__
        lbl = tk.Label(frame1, text=description)
        lbl.grid(pady="5", row=1, columnspan=4)

        ###########
        # Frame 2 #
        ###########

        frame2 = tk.LabelFrame(self.master, text="Program Console")
        frame2.grid(padx=5, pady=5, ipadx=2, ipady=2, sticky=tk.N, row=1, column=0)

        btn = ttk.Button(frame2, text="Browse file", command=self.open_file)
        btn.config(width=14)
        btn.grid(padx="10", pady="5", row=0)

        btn = ttk.Button(frame2, text="Compute", command=self.calc_distortion)
        btn.config(width=14)
        btn.grid(padx="10", pady="5", row=1)

        btn = ttk.Button(frame2, text="Clear cache", command=self.clear_cache)
        btn.config(width=14)
        btn.grid(padx="10", pady="5", row=2)

        btn = ttk.Button(frame2, text="Save Results", command=self.save_results)
        btn.config(width=14)
        btn.grid(padx="10", pady="5", row=3)

        ###########
        # Frame 3 #
        ###########

        frame3 = tk.LabelFrame(self.master, text="Distortion Parameters")
        frame3.grid(padx=5, pady=5, ipadx=3, ipady=2, sticky=tk.N, row=1, column=1)

        # D_mean
        lbl = tk.Label(frame3, text="<D>   =   ")
        lbl.grid(sticky=tk.E, pady="5", row=0, column=0)

        self.box_d_mean = tk.Entry(frame3)
        self.box_d_mean.configure(width="12", justify='center')
        self.box_d_mean.grid(row=0, column=1)

        lbl = tk.Label(frame3, text="  Angstrom")
        lbl.grid(pady="5", row=0, column=2)

        # Zeta
        lbl = tk.Label(frame3, text="ζ   =   ")
        lbl.grid(sticky=tk.E, pady="5", row=1, column=0)

        self.box_zeta = tk.Entry(frame3)
        self.box_zeta.configure(width="12", justify='center')
        self.box_zeta.grid(row=1, column=1)

        lbl = tk.Label(frame3, text="  Angstrom")
        lbl.grid(pady="5", row=1, column=2)

        # Delta
        lbl_delta = tk.Label(frame3, text="Δ   =   ")
        lbl_delta.grid(sticky=tk.E, pady="5", row=2, column=0)

        self.box_delta = tk.Entry(frame3)
        self.box_delta.configure(width="12", justify='center')
        self.box_delta.grid(row=2, column=1)

        # Sigma
        lbl = tk.Label(frame3, text="Σ   =   ")
        lbl.grid(sticky=tk.E, pady="5", row=3, column=0)

        self.box_sigma = tk.Entry(frame3)
        self.box_sigma.configure(width="12", justify='center')
        self.box_sigma.grid(row=3, column=1)

        lbl = tk.Label(frame3, text="  degree")
        lbl.grid(pady="5", row=3, column=2)

        # Theta_mean
        lbl = tk.Label(frame3, text="Θ   =   ")
        lbl.grid(sticky=tk.E, pady="5", row=4, column=0)

        self.box_theta_mean = tk.Entry(frame3)
        self.box_theta_mean.configure(width="12", justify='center')
        self.box_theta_mean.grid(row=4, column=1)

        lbl = tk.Label(frame3, text="  degree")
        lbl.grid(pady="5", row=4, column=2)

        ###########
        # Frame 4 #
        ###########

        frame4 = tk.Frame(self.master)
        frame4.grid(padx=5, pady=10, row=2, column=0, columnspan=2)

        self.box_result = tkscrolled.ScrolledText(frame4)
        self.box_result.configure(height="19", width="70", wrap="word", undo="True")
        self.box_result.grid(row=0)

    def show_text(self, text):
        """
        Insert text to result box

        Parameters
        ----------
        text : str
            Text to show in result box.

        Returns
        -------
        None : None

        """
        self.box_result.insert(tk.INSERT, text + "\n")
        self.box_result.see(tk.END)

    def welcome_msg(self):
        """
        Show welcome message in result box:

        1. Program name, version, and release.
        2. Full author names.
        3. Official website: https://octadist.github.io.

        """
        full_version = octadist.__version__ + " " + octadist.__release__

        self.show_text(f"Welcome to OctaDist {full_version}\n")
        self.show_text(f"Developed by {octadist.__author_full__}.\n")
        self.show_text(octadist.__website__ + "\n")

    #####################
    # Manipulating File #
    #####################

    def open_file(self):
        """
        Open file dialog where the user will browse input files.

        """
        self.clear_cache()

        input_file = filedialog.askopenfilenames(
            title="Choose input file",
            filetypes=(("XYZ File", "*.xyz"),
                       ("Gaussian Output File", "*.out"),
                       ("Gaussian Output File", "*.log"),
                       ("NWChem Output File", "*.out"),
                       ("NWChem Output File", "*.log"),
                       ("ORCA Output File", "*.out"),
                       ("ORCA Output File", "*.log"),
                       ("Q-Chem Output File", "*.out"),
                       ("Q-Chem Output File", "*.log"),
                       ("All Files", "*.*")))

        self.file_list = list(input_file)

        self.search_coord()

    def search_coord(self):
        """
        Search and extract atomic symbols and coordinates from input file.

        See Also
        --------
        octadist.src.molecule.extract_coord :
            Extract atomic symbols and atomic coordinates from input file.
        octadist.src.molecule.extract_octa :
            Extract octahedral structure from complex.

        """
        try:
            open(self.file_list[0], 'r')
        except IndexError:
            return 1

        for i in range(len(self.file_list)):

            ########################################
            # Extract atomic coordinates from file #
            ########################################

            atom_full, coord_full = molecule.extract_coord(self.file_list[i])
            self.atom_coord_full.append([atom_full, coord_full])

            # If either lists is empty, then continue to next file
            if len(list(atom_full)) == 0 or len(coord_full) == 0:
                continue

            #################################################
            # Extract octahedral structure from the complex #
            #################################################

            total_metal, atom_metal, coord_metal = molecule.find_metal(atom_full, coord_full)

            if total_metal == 0:
                popup.warn_no_metal(i + 1)

            # loop over number of metal found in complex
            for j in range(total_metal):
                atom_octa, coord_octa = molecule.extract_octa(atom_full,
                                                              coord_full,
                                                              j + 1,
                                                              self.cutoff_metal_ligand)

                # If no atomic coordinates inside, it will raise error
                if np.any(coord_octa) == 0:
                    popup.err_no_coord(i + 1)
                    continue

                if len(coord_octa) < 7:
                    self.clear_result_box()
                    popup.err_less_ligands(i + 1)
                    continue

                # File number and file name
                file_name = self.file_list[i].split('/')[-1]
                self.file_name.append([i + 1, file_name])

                # Metal center atom
                self.octa_index.append(atom_octa[0])

                # Atomic labels and atomic coordinates
                # Example:
                #
                # atom = ['Fe', 'O', 'O', 'N', 'N', 'N', 'N']
                # coord = [[2.298354, 5.161785, 7.971898],
                #          [1.885657, 4.804777, 6.183726],
                #          [1.747515, 6.960963, 7.932784],
                #          [4.094380, 5.807257, 7.588689],
                #          [0.539005, 4.482809, 8.460004],
                #          [2.812425, 3.266553, 8.131637],
                #          [2.886404, 5.392925, 9.848966]]

                self.atom_coord_octa.append([atom_octa, coord_octa])

        self.show_coord()

    def show_coord(self):
        """
        Show coordinates in box.

        """
        # loop over complex
        for i in range(len(self.atom_coord_octa)):
            if i == 0:
                self.show_text("XYZ coordinates of extracted octahedral structure")

            self.show_text(f"File {self.file_name[i][0]}: {self.file_name[i][1]}")
            self.show_text(f"Metal center atom: {self.octa_index[i]}")
            self.show_text("Atom                       Cartesian coordinate")

            # loop over atoms in octahedron
            for k in range(7):
                self.show_text(" {0:>2}      {1:14.9f}  {2:14.9f}  {3:14.9f}"
                               .format(self.atom_coord_octa[i][0][k],
                                       self.atom_coord_octa[i][1][k][0],
                                       self.atom_coord_octa[i][1][k][1],
                                       self.atom_coord_octa[i][1][k][2])
                               )
            self.show_text("")

    def save_results(self):
        """
        Save results as output file. Default file extension is *.txt.

        """
        f = filedialog.asksaveasfile(mode='w', defaultextension=".txt", title="Save results",
                                     filetypes=(("TXT File", "*.txt"),
                                                ("All Files", "*.*")))

        if f is None:
            return 0

        f.write("OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.\n")
        f.write("===========================================================\n")
        f.write("\n")
        f.write(f"OctaDist {octadist.__version__} {octadist.__release__}.\n")
        f.write("Octahedral Distortion Calculator\n")
        f.write(f"{octadist.__website__}\n")
        f.write("\n")
        f.write("================ Start of the Output file =================\n")
        f.write("\n")
        get_result = self.box_result.get('1.0', tk.END + '-1c')
        f.write(get_result)
        f.write("\n")
        f.write("================= End of the output file ==================\n")
        f.write("\n")
        f.close()

        popup.info_save_results(f.name)

    def save_coord(self):
        """
        Save atomic coordinates (Cartesian coordinate) of octahedral structure.
        Default file extension is *xyz.

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        if len(self.file_list) > 1:
            popup.err_many_files()
            return 1

        f = filedialog.asksaveasfile(mode='w', defaultextension=".xyz",
                                     title="Save atomic coordinates",
                                     filetypes=(("XYZ File", "*.xyz"),
                                                ("TXT File", "*.txt"),
                                                ("All Files", "*.*")))

        file_name = self.file_list[0].split('/')[-1]
        atoms = self.atom_coord_octa[0][0]
        coord = self.atom_coord_octa[0][1]

        full_version = octadist.__version__ + " " + octadist.__release__

        f.write("7\n")
        f.write(f"{file_name} : this file was generated by OctaDist {full_version}.\n")
        for i in range(7):
            f.write("{0:2s}\t{1:9.6f}\t{2:9.6f}\t{3:9.6f}\n"
                    .format(atoms[i], coord[i][0], coord[i][1], coord[i][2]))
        f.write("\n")
        f.close()

        popup.info_save_results(f.name)

    ###################################
    # Calculate distortion parameters #
    ###################################

    def calc_distortion(self):
        """
        Calculate all distortion parameters:

        - D_mean
        - Zeta
        - Delta
        - Sigma
        - Theta

        See Also
        --------
        octadist.src.calc.CalcDistortion.calc_d_mean :
            Calculate mean metal-ligand bond length.
        octadist.src.calc.CalcDistortion.calc_zeta :
            Calculate Zeta parameter.
        octadist.src.calc.CalcDistortion.calc_delta :
            Calculate Delta parameter.
        octadist.src.calc.CalcDistortion.calc_sigma :
            Calculate Sigma parameter.
        octadist.src.calc.CalcDistortion.calc_theta :
            Calculate Theta parameter.

        """
        if len(self.atom_coord_octa) >= 1:
            self.clear_param_box()
        else:
            popup.err_no_file()
            return 1

        # if comp_result is not empty, clean it to avoid over loop.
        if self.comp_result:
            self.comp_result = []

        # loop over number of metal complexes
        for i in range(len(self.atom_coord_octa)):
            atom_octa, coord_octa = self.atom_coord_octa[i]

            # Calculate distortion parameters
            calc_dist = calc.CalcDistortion(coord_octa)

            d_mean = calc_dist.d_mean
            zeta = calc_dist.zeta
            delta = calc_dist.delta
            sigma = calc_dist.sigma
            theta = calc_dist.theta
            non_octa = calc_dist.non_octa

            if non_octa:
                popup.warn_not_octa()

            # Collect results
            self.all_zeta.append(zeta)
            self.all_delta.append(delta)
            self.all_sigma.append(sigma)
            self.all_theta.append(theta)

            self.comp_result.append([d_mean, zeta, delta, sigma, theta])

        # Print results to each unique box.

        if len(self.atom_coord_octa) == 1:
            d_mean, zeta, delta, sigma, theta = self.comp_result[0]

            self.box_d_mean.insert(tk.INSERT, f"{d_mean:3.6f}")
            self.box_zeta.insert(tk.INSERT, f"{zeta:3.6f}")
            self.box_delta.insert(tk.INSERT, f"{delta:3.6f}")
            self.box_sigma.insert(tk.INSERT, f"{sigma:3.6f}")
            self.box_theta_mean.insert(tk.INSERT, f"{theta:3.6f}")
        else:
            self.box_d_mean.insert(tk.INSERT, "See below")
            self.box_zeta.insert(tk.INSERT, "See below")
            self.box_delta.insert(tk.INSERT, "See below")
            self.box_sigma.insert(tk.INSERT, "See below")
            self.box_theta_mean.insert(tk.INSERT, "See below")

        # Print results to result box
        self.show_text("Computed octahedral distortion parameters for all complexes\n")
        self.show_text("No. - Metal\t\tD_mean\tZeta\tDelta\tSigma\tTheta")
        self.show_text("**********************************************************************\n")
        for i in range(len(self.comp_result)):
            self.show_text("{0:2d}  -  {1}\t\t{2:9.4f}\t{3:9.6f}\t{4:9.6f}\t{5:9.4f}\t{6:9.4f}"
                           .format(i + 1,
                                   self.octa_index[i],
                                   self.comp_result[i][0],
                                   self.comp_result[i][1],
                                   self.comp_result[i][2],
                                   self.comp_result[i][3],
                                   self.comp_result[i][4])
                           )

        self.show_text("")

    ###################
    # Program Setting #
    ###################

    def settings(self):
        """
        Program settings allows the user to configure the values of variables
        that used in molecular display function.

        For example, cutoff distance for screening bond distance between atoms.

        """

        def open_exe():
            """
            Program setting: Open dialog in which the user will choose text editor.

            """
            try:
                input_file = filedialog.askopenfilename(
                    title="Choose text editor executable",
                    filetypes=[("EXE file", "*.exe")]
                )

                file_list = str(input_file)
                entry_exe.delete(0, tk.END)
                entry_exe.insert(tk.INSERT, file_list)

            except IndexError:
                return 1

        def is_show_title():
            """
            Check if title of figure will be set to show or not.

            """
            if var_title.get():
                var_title.set(True)
            else:
                var_title.set(False)

        def is_show_axis():
            """
            Check if axis of figure will be set to show or not.

            """
            if var_axis.get():
                var_axis.set(True)
            else:
                var_axis.set(False)

        def is_show_grid():
            """
            Check if grid of figure will be set to show or not.

            """
            if var_grid.get():
                var_grid.set(True)
            else:
                var_grid.set(False)

        def restore_settings(self):
            """
            Restore all settings.

            """
            self.cutoff_metal_ligand = self.backup_cutoff_metal_ligand
            self.cutoff_global = self.backup_cutoff_global
            self.cutoff_hydrogen = self.backup_cutoff_hydrogen
            self.text_editor = self.backup_text_editor
            self.show_title = self.backup_show_title
            self.show_axis = self.backup_show_axis
            self.show_grid = self.backup_show_grid

            var_1.set(self.cutoff_metal_ligand)
            var_2.set(self.cutoff_global)
            var_3.set(self.cutoff_hydrogen)
            var_title.set(self.show_title)
            var_axis.set(self.show_axis)
            var_grid.set(self.show_grid)

            entry_exe.delete(0, tk.END)
            entry_exe.insert(tk.INSERT, self.text_editor)

        def click_ok(self):
            """
            If the user click OK, it will save all settings and show info in output box.

            """
            self.cutoff_metal_ligand = float(var_1.get())
            self.cutoff_global = float(var_2.get())
            self.cutoff_hydrogen = float(var_3.get())
            self.text_editor = str(entry_exe.get())
            self.show_title = bool(var_title.get())
            self.show_axis = bool(var_axis.get())
            self.show_grid = bool(var_grid.get())

            self.show_text("Updated program settings")
            self.show_text("************************")
            self.show_text(f"Metal-ligand bond cutoff\t\t\t{self.cutoff_metal_ligand}")
            self.show_text(f"Global bond cutoff\t\t\t{self.cutoff_global}")
            self.show_text(f"Hydrogen bond cutoff\t\t\t{self.cutoff_hydrogen}")
            self.show_text(f"Text editor\t\t\t{self.text_editor}")
            self.show_text(f"Show Title\t\t\t{self.show_title}")
            self.show_text(f"Show Axis\t\t\t{self.show_axis}")
            self.show_text(f"Show Grid\t\t\t{self.show_grid}\n")

            wd.destroy()

        def click_cancel():
            """
            If the user click CANCEL, close window.

            """
            wd.destroy()

        ###################
        # Setting: Widget #
        ###################

        wd = tk.Toplevel(self.master)
        if self.octadist_icon is not None:
            wd.wm_iconbitmap(self.octadist_icon)
        wd.title("Program settings")
        wd.option_add("*Font", "Arial 10")

        frame = tk.Frame(wd)
        frame.grid()

        ###################
        # Setting: Cutoff #
        ###################

        cutoff = tk.LabelFrame(frame, text="Bond Cutoff:")
        cutoff.grid(padx=5, pady=5, ipadx=5, ipady=5, sticky='W', row=0, columnspan=4)

        label_1 = tk.Label(cutoff, text="Metal-Ligand Bond")
        label_1.grid(padx="10", pady="5", ipadx="10", row=0, column=0)

        var_1 = tk.DoubleVar()
        var_1.set(self.cutoff_metal_ligand)

        scale_1 = tk.Scale(cutoff, orient="horizontal", variable=var_1, to=5, resolution=0.1)
        scale_1.configure(width=20, length=100)
        scale_1.grid(padx="10", pady="5", ipadx="10", row=1, column=0)

        label_2 = tk.Label(cutoff, text="Global Distance")
        label_2.grid(padx="10", pady="5", ipadx="10", row=0, column=1)

        var_2 = tk.DoubleVar()
        var_2.set(self.cutoff_global)

        scale_2 = tk.Scale(cutoff, orient="horizontal", variable=var_2, to=5, resolution=0.1)
        scale_2.configure(width=20, length=100)
        scale_2.grid(padx="10", pady="5", ipadx="10", row=1, column=1)

        label_3 = tk.Label(cutoff, text="Hydrogen Distance")
        label_3.grid(padx="10", pady="5", ipadx="10", row=0, column=2)

        var_3 = tk.DoubleVar()
        var_3.set(self.cutoff_hydrogen)

        scale_3 = tk.Scale(cutoff, orient="horizontal", variable=var_3, to=5, resolution=0.1)
        scale_3.configure(width=20, length=100)
        scale_3.grid(padx="10", pady="5", ipadx="10", row=1, column=2)

        ########################
        # Setting: Text editor #
        ########################

        frame_text_editor = tk.LabelFrame(frame, text="Text editor:")
        frame_text_editor.grid(padx=5, pady=5, ipadx=5, ipady=5, sticky='W', row=1, columnspan=4)

        label = tk.Label(frame_text_editor, text="Enter the EXE:")
        label.grid(padx="5", sticky=tk.E, row=0, column=0)

        entry_exe = tk.Entry(frame_text_editor, bd=2, width=60)
        entry_exe.grid(row=0, column=1)

        button = tk.Button(frame_text_editor, text="Browse...", command=open_exe)
        button.grid(padx="5", pady="5", sticky=tk.W, row=0, column=2)

        entry_exe.insert(tk.INSERT, self.text_editor)

        #####################
        # Setting: Displays #
        #####################

        displays = tk.LabelFrame(frame, text="Displays:")
        displays.grid(padx=5, pady=5, ipadx=5, ipady=5, sticky='W', row=2, columnspan=4)

        # Show title of plot?
        var_title = tk.BooleanVar()
        var_title.set(self.show_title)

        show_title = ttk.Checkbutton(displays, text="Show Figure Title", onvalue=True, offvalue=False,
                                     variable=var_title, command=lambda: is_show_title())
        show_title.grid(padx="5", pady="5", ipadx="25", sticky=tk.E, row=0, column=0)

        # Show axis?
        var_axis = tk.BooleanVar()
        var_axis.set(self.show_axis)

        show_axis = ttk.Checkbutton(displays, text="Show Axis", onvalue=True, offvalue=False,
                                    variable=var_axis, command=lambda: is_show_axis())
        show_axis.grid(padx="5", pady="5", ipadx="25", sticky=tk.E, row=0, column=1)

        # Show grid?
        var_grid = tk.BooleanVar()
        var_grid.set(self.show_grid)

        show_grid = ttk.Checkbutton(displays, text="Show Gridlines", onvalue=True, offvalue=False,
                                    variable=var_grid, command=lambda: is_show_grid())
        show_grid.grid(padx="5", pady="5", ipadx="5", sticky=tk.E, row=0, column=2)

        ####################
        # Setting: Console #
        ####################

        button = tk.Button(frame, text="Restore settings", command=lambda: restore_settings(self))
        button.configure(width=15)
        button.grid(padx="10", pady="10", sticky=tk.W, row=3, column=0)

        button = tk.Button(frame, text="OK", command=lambda: click_ok(self))
        button.configure(width=15)
        button.grid(padx="5", pady="10", sticky=tk.E, row=3, column=2)

        button = tk.Button(frame, text="Cancel", command=lambda: click_cancel())
        button.configure(width=15)
        button.grid(padx="5", pady="10", row=3, column=3)

        frame.mainloop()

    #################
    # Copy and Edit #
    #################

    def copy_name(self):
        """
        Copy input file name to clipboard.

        See Also
        --------
        copy_path
        copy_results
        copy_octa

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        name = self.file_list[0].split('/')[-1]

        clip = tk.Tk()
        clip.withdraw()
        clip.clipboard_clear()
        clip.clipboard_append(name)
        clip.destroy()

    def copy_path(self):
        """
        Copy absolute path of input file to clipboard.

        See Also
        --------
        copy_name
        copy_results
        copy_octa

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        clip = tk.Tk()
        clip.withdraw()
        clip.clipboard_clear()
        clip.clipboard_append(self.file_list[0])
        clip.destroy()

    def copy_results(self):
        """
        Copy the results and computed distortion parameters to clipboard.

        See Also
        --------
        copy_name
        copy_path
        copy_octa

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        if len(self.all_zeta) == 0:
            popup.err_no_calc()
            return 1

        results = "Zeta, Delta, Sigma, Gamma\n" \
                  "{0:3.6f}, {1:3.6f}, {2:3.6f}, {3:3.6f}" \
            .format(self.all_zeta[0], self.all_delta[0], self.all_sigma[0], self.all_theta[0])

        clip = tk.Tk()
        clip.withdraw()
        clip.clipboard_clear()
        clip.clipboard_append(results)
        clip.destroy()

    def copy_octa(self):
        """
        Copy atomic coordinates of octahedral structure to clipboard.

        See Also
        --------
        copy_name
        copy_path
        copy_results

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        clip = tk.Tk()
        clip.withdraw()
        clip.clipboard_clear()
        clip.clipboard_append(self.atom_coord_octa[0][2])
        clip.destroy()

    def edit_file(self):
        """
        Edit file by specified text editor on Windows.

        See Also
        --------
        settings

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        if self.text_editor == "":
            popup.err_no_editor()
            return 1

        try:
            for i in range(len(self.file_list)):
                program_name = self.text_editor
                file_name = self.file_list[i]
                subprocess.Popen([program_name, file_name])

        except FileNotFoundError:
            return 1

    #########################
    # Interactive scripting #
    #########################

    def scripting_console(self):
        """
        Start scripting interface for an interactive code.

        User can access to class variable (dynamics variable).

        +------------+
        | Output box |
        +------------+
        | Input box  |
        +------------+

        See Also
        --------
        settings :
            Program settings.

        """
        my_app = scripting.ScriptingConsole(self)
        my_app.scripting_start()

    ##################
    # Visualizations #
    ##################

    def draw_all_atom(self):
        """
        Display 3D complex.

        See Also
        --------
        octadist.src.draw.DrawComplex :
            Show 3D molecule.

        """
        if len(self.atom_coord_full) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_full) > 1:
            popup.err_many_files()
            return 1

        atom_full, coord_full = self.atom_coord_full[0]

        my_plot = draw.DrawComplex(atom=atom_full,
                                   coord=coord_full,
                                   cutoff_global=self.cutoff_global,
                                   cutoff_hydrogen=self.cutoff_hydrogen)
        my_plot.add_atom()
        my_plot.add_bond()
        my_plot.add_legend()
        my_plot.config_plot(show_title=self.show_title,
                            show_axis=self.show_axis,
                            show_grid=self.show_grid)
        my_plot.show_plot()

    def draw_all_atom_and_face(self):
        """
        Display 3D complex with the faces.

        See Also
        --------
        octadist.src.draw.DrawComplex :
            Show 3D molecule.

        """
        if len(self.atom_coord_full) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_full) > 1:
            popup.err_many_files()
            return 1

        atom_full, coord_full = self.atom_coord_full[0]

        my_plot = draw.DrawComplex(atom=atom_full,
                                   coord=coord_full,
                                   cutoff_global=self.cutoff_global,
                                   cutoff_hydrogen=self.cutoff_hydrogen)
        my_plot.add_atom()
        my_plot.add_bond()

        for i in range(len(self.atom_coord_octa)):
            _, coord_octa = self.atom_coord_octa[i]
            my_plot.add_face(coord_octa)

        my_plot.add_legend()
        my_plot.config_plot(show_title=self.show_title,
                            show_axis=self.show_axis,
                            show_grid=self.show_grid)
        my_plot.show_plot()

    def draw_octa(self):
        """
        Display 3D octahedral structure.

        See Also
        --------
        octadist.src.draw.DrawComplex :
            Show 3D molecule.

        """
        if len(self.atom_coord_octa) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_octa) > 1:
            popup.err_many_files()
            return 1

        atom_octa, coord_octa = self.atom_coord_octa[0]

        my_plot = draw.DrawComplex(atom=atom_octa,
                                   coord=coord_octa,
                                   cutoff_global=self.cutoff_global,
                                   cutoff_hydrogen=self.cutoff_hydrogen)
        my_plot.add_atom()
        my_plot.add_bond()
        my_plot.add_legend()
        my_plot.config_plot(show_title=self.show_title,
                            show_axis=self.show_axis,
                            show_grid=self.show_grid)
        my_plot.show_plot()

    def draw_octa_and_face(self):
        """
        Display 3D octahedral structure with the faces.

        See Also
        --------
        octadist.src.draw.DrawComplex :
            Show 3D molecule.

        """
        if len(self.atom_coord_octa) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_octa) > 1:
            popup.err_many_files()
            return 1

        atom_octa, coord_octa = self.atom_coord_octa[0]

        my_plot = draw.DrawComplex(atom=atom_octa,
                                   coord=coord_octa,
                                   cutoff_global=self.cutoff_global,
                                   cutoff_hydrogen=self.cutoff_hydrogen)
        my_plot.add_atom()
        my_plot.add_bond()
        my_plot.add_legend()

        for i in range(len(self.atom_coord_octa)):
            _, coord = self.atom_coord_octa[i]
            my_plot.add_face(coord)

        my_plot.config_plot(show_title=self.show_title,
                            show_axis=self.show_axis,
                            show_grid=self.show_grid)
        my_plot.show_plot()

    def draw_projection(self):
        """
        Draw projection planes.

        See Also
        --------
        octadist.src.draw.DrawProjection :
            Show graphical projections.

        """
        if len(self.atom_coord_full) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_full) > 1:
            popup.err_many_files()
            return 1

        atom_full, coord_full = self.atom_coord_full[0]

        my_plot = draw.DrawProjection(atom=atom_full, coord=coord_full)
        my_plot.add_atom()
        my_plot.add_symbol()
        my_plot.add_plane()
        my_plot.show_plot()

    def draw_twisting_plane(self):
        """
        Draw twisting triangular planes.

        See Also
        --------
        octadist.src.draw.DrawTwistingPlane :
            Show graphical triangular twisting planes.

        """
        if len(self.atom_coord_full) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_full) > 1:
            popup.err_many_files()
            return 1

        atom_full, coord_full = self.atom_coord_full[0]

        my_plot = draw.DrawTwistingPlane(atom=atom_full, coord=coord_full)
        my_plot.add_plane()
        my_plot.add_symbol()
        my_plot.add_bond()
        my_plot.show_plot()

    #####################
    # Show data summary #
    #####################

    def show_data_complex(self):
        """
        Show info of input complex.

        See Also
        --------
        octadist.src.structure.DataComplex :
            Show data summary of complex.

        """
        if len(self.file_list) == 0:
            popup.err_no_file()
            return 1

        my_app = structure.DataComplex(master=self.master, icon=self.octadist_icon)

        for i in range(len(self.file_list)):
            atom = self.atom_coord_full[i][0]
            coord = self.atom_coord_full[i][1]
            my_app.add_name(self.file_list[i])
            my_app.add_coord(atom, coord)

    def show_param_octa(self):
        """
        Show structural parameters of selected octahedral structure.

        See Also
        --------
        octadist.src.structure.StructParam :
            Show structural parameter symmary of complex.

        """
        if len(self.atom_coord_octa) == 0:
            popup.err_no_file()
            return 1

        my_app = structure.StructParam(master=self.master, icon=self.octadist_icon)

        for i in range(len(self.atom_coord_octa)):
            metal = self.octa_index[i]
            atom, coord = self.atom_coord_octa[i]
            my_app.add_metal(metal)
            my_app.add_coord(atom, coord)

    def show_surface_area(self):
        """
        Calculate the area of eight triangular faces of octahedral structure.

        See Also
        --------
        octadist.src.structure.SurfaceArea :
            Show the area of the faces of octahedral structure.

        """
        if len(self.atom_coord_octa) == 0:
            popup.err_no_file()
            return 1

        my_app = structure.SurfaceArea(master=self.master, icon=self.octadist_icon)

        for i in range(len(self.atom_coord_octa)):
            metal = self.octa_index[i]
            atom, coord = self.atom_coord_octa[i]
            my_app.add_metal(metal)
            my_app.add_octa(coord)

    ##############################
    # Plot between two data sets #
    ##############################

    def plot_zeta_sigma(self):
        """
        Plot relationship between zeta and sigma.

        See Also
        --------
        octadist.src.plot.Plot :
            Show relationship plot.

        """
        if len(self.all_sigma) == 0:
            popup.err_no_calc()
            return 1

        my_plot = plot.Plot(self.all_zeta, self.all_sigma, name1="zeta", name2="sigma")
        my_plot.add_point()
        my_plot.add_text()
        my_plot.add_legend()
        my_plot.show_plot()

    def plot_sigma_theta(self):
        """
        Plot relationship between sigma and theta.

        See Also
        --------
        octadist.src.plot.Plot :
            Show relationship plot.

        """
        if len(self.all_sigma) == 0:
            popup.err_no_calc()
            return 1

        my_plot = plot.Plot(self.all_sigma, self.all_theta, name1="sigma", name2="theta")
        my_plot.add_point()
        my_plot.add_text()
        my_plot.add_legend()
        my_plot.show_plot()

    ##################
    # Analysis tools #
    ##################

    def tool_jahn_teller(self):
        """
        Calculate Jahn-Teller distortion parameter.

        See Also
        --------
        octadist.src.tools.CalcJahnTeller :
            Calculate Jahn-Teller distortion parameter.

        """
        if len(self.atom_coord_full) == 0:
            popup.err_no_file()
            return 1
        elif len(self.atom_coord_full) > 1:
            popup.err_many_files()
            return 1

        atom_full, coord_full = self.atom_coord_full[0]

        run_jt = tools.CalcJahnTeller(atom=atom_full,
                                      coord=coord_full,
                                      cutoff_global=self.cutoff_global,
                                      cutoff_hydrogen=self.cutoff_hydrogen,
                                      master=self.master,
                                      icon=self.octadist_icon)
        run_jt.start_app()
        run_jt.find_bond()
        run_jt.show_app()

    def tool_rmsd(self):
        """
        Calculate root mean squared displacement of atoms in complex, RMSD.

        See Also
        --------
        octadist.src.tools.CalcRMSD :
            Calculate RMSD.

        """
        if len(self.atom_coord_full) != 2:
            popup.err_only_2_files()
            return 1

        complex_1 = self.atom_coord_full[0]
        complex_2 = self.atom_coord_full[1]

        atom_complex_1, coord_complex_1 = complex_1
        atom_complex_2, coord_complex_2 = complex_2

        # Check if two complexes are consistent
        if len(atom_complex_1) != len(atom_complex_2):
            popup.err_not_equal_atom()
            return 1

        for i in range(len(atom_complex_1)):
            if atom_complex_1[i] != atom_complex_2[i]:
                popup.err_atom_not_match(i + 1)
                return 1

        run_rmsd = tools.CalcRMSD(coord_complex_1, coord_complex_2,
                                  atom_1=atom_complex_1, atom_2=atom_complex_2,
                                  master=self.master, icon=self.octadist_icon)

        run_rmsd.start_app()
        run_rmsd.show_app()

    ################
    # Check Update #
    ################

    @staticmethod
    def check_update():
        """
        Check program update by comparing version of program user is using with
        that of the latest version released on github.

        References
        ----------
        File: https://www.github.com/OctaDist/OctaDist/version_update.txt.

        """
        data = urlopen("https://raw.githubusercontent.com/OctaDist/OctaDist/master/version_update.txt").read()
        # decode
        data = data.decode('utf-8')
        data = data.split()

        user_rev = float(octadist.__revision__)
        server_ver = data[1]
        server_rev = float(data[3])  # code version

        os_name = platform.system()  # find the OS name

        if server_rev > user_rev:
            popup.info_new_update()

            text = f"A new version {server_ver} is ready for download.\n\n" \
                "Do you want to download now?"
            msg_box = messagebox.askquestion("Updates available", text, icon="warning")

            if msg_box == 'yes':

                dl_link = "https://github.com/OctaDist/OctaDist/releases/download/"
                main_link = dl_link + "v." + server_ver + "/OctaDist-" + server_ver

                if os_name == "Windows":
                    link_windows = main_link + "-Win-x86-64.exe"
                    webbrowser.open_new_tab(link_windows)

                elif os_name == "Darwin" or os_name == "Linux":
                    link_linux = main_link + "-src-x86-64.tar.gz"
                    webbrowser.open_new_tab(link_linux)

                else:
                    popup.err_cannot_update()

                # Open Thank You page at the same time download the program
                webbrowser.open_new_tab("https://octadist.github.io/thanks.html")

            else:
                pass

        elif server_rev < user_rev:
            popup.info_using_dev()

        else:
            popup.info_no_update()

    #####################
    # Show program info #
    #####################

    @staticmethod
    def callback(event):
        """
        On-clink open web browser.

        Parameters
        ----------
        event : object
            Event object for callback.

        """
        webbrowser.open_new(event.widget.cget("text"))

    def show_help(self):
        """
        Show program help on a sub-window:

        1. Simple usage
        2. XYZ file format
        3. References

        """
        wd = tk.Toplevel(self.master)
        if self.octadist_icon is not None:
            wd.wm_iconbitmap(self.octadist_icon)
        wd.title("Program Help")
        wd.option_add("*Font", "Arial 10")
        wd.geometry("550x420")
        wd.resizable(0, 0)
        frame = tk.Frame(wd)
        frame.grid()

        # Usage
        lbl = tk.Label(frame, text="Usage:")
        lbl.grid(sticky=tk.W, row=0)

        msg_help_1 = """\
1. Browse input file
2. Check if uploaded data is accurate
3. Compute distortion parameters
4. Analyse and refine results
5. Save results as output file
"""

        msg = tk.Message(frame, text=msg_help_1, width="450")
        msg.grid(padx="15", sticky=tk.W, row=1)

        # XYZ file format
        lbl = tk.Label(frame, text="Input structure in .xyz format:\n")
        lbl.grid(sticky=tk.W, row=2)

        msg_help_2 = """\
1)\t<number of atoms>
2)\tcomment line
3)\t<Atom 1>\t<X>\t<Y>\t<Z>\t
4)\t<Atom 2>\t<X>\t<Y>\t<Z>\t
5)\t<Atom 3>\t<X>\t<Y>\t<Z>\t
6)\t<Atom 4>\t<X>\t<Y>\t<Z>\t
.....
.....
n)\t<Atom n>\t<X>\t<Y>\t<Z>\t
"""

        msg = tk.Message(frame, text=msg_help_2, width="450", bg="yellow")
        msg.grid(padx="15", sticky=tk.W, row=3, column=0)

        lbl = tk.Label(frame, text="\nOther examples of input file:")
        lbl.grid(sticky=tk.W, row=5, columnspan=2)
        link = "https://github.com/OctaDist/OctaDist/tree/master/example-input\n"
        lbl_link = tk.Label(frame, foreground="blue", text=link, cursor="hand2")
        lbl_link.grid(padx="15", pady="5", sticky=tk.W, row=6, columnspan=2)
        lbl_link.bind("<Button-1>", self.callback)

    @staticmethod
    def show_about():
        """
        Show author details on a sub-window.

        1. Name of authors
        2. Official program website
        3. Citation

        """
        text = f"OctaDist version {octadist.__version__} ({octadist.__release__})\n\n" \
            f"Authors: {octadist.__author_full__}.\n\n" \
            f"Website: {octadist.__website__}\n\n" \
            f"Please cite this project if you use OctaDist for scientific publication."

        showinfo("About program", text)

    @staticmethod
    def show_license():
        """
        Show license details on a sub-window.

        GNU General Public License version 3.0.

        References
        ----------
        Link: https://www.gnu.org/licenses/gpl-3.0.en.html.

        """
        text = """\
OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.

This program is free software: you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation, either version 3 of the License, or \
(at your option) any later version.

This program is distributed in the hope that it will be useful, \
but WITHOUT ANY WARRANTY; without even the implied warranty of \
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \
GNU General Public License for more details.

You should have received a copy of the GNU General Public License \
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

        showinfo("License", text)

    ###################
    # Clear All Cache #
    ###################

    def clear_cache(self):
        """
        Clear program cache by nullifying all default variables
        and clear both of parameter and result boxes.

        """
        for name in dir():
            if not name.startswith('_'):
                del locals()[name]

        self.file_list = []
        self.file_name = []
        self.octa_index = []
        self.atom_coord_full = []
        self.atom_coord_octa = []
        self.all_zeta = []
        self.all_delta = []
        self.all_sigma = []
        self.all_theta = []
        self.comp_result = []

        self.clear_param_box()
        self.clear_result_box()

    def clear_param_box(self):
        """
        Clear parameter box.

        """
        self.box_delta.delete(0, tk.END)
        self.box_sigma.delete(0, tk.END)
        self.box_d_mean.delete(0, tk.END)
        self.box_zeta.delete(0, tk.END)
        self.box_theta_mean.delete(0, tk.END)

    def clear_result_box(self):
        """
        Clear result box.

        """
        self.box_result.delete(1.0, tk.END)

    def start_app(self):
        """
        Start application.

        """
        self.master.mainloop()


def main():
    app = OctaDist()
    app.start_app()

    # Delete icon after closing app
    if app.octadist_icon is not None:
        os.remove(app.octadist_icon)


if __name__ == '__main__':
    main()



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import functools
import scipy.optimize

import numpy as np


def find_eq_of_plane(x, y, z):
    """
    Find the equation of plane of given three points using cross product

    | The general form of plane equation:
    |
    | Ax + By + Cz = D
    |
    | where A, B, C, and D are coefficient.
    |
    | XZ  X  XY = (a, b, c)
    |
    | d = (a, b, c).Z

    Parameters
    ----------
    x : array_like
        3D Coordinate of point.
    y : array_like
        3D Coordinate of point.
    z : array_like
        3D Coordinate of point.

    Returns
    -------
    a : float64
        Coefficient of the equation of the plane.
    b : float64
        Coefficient of the equation of the plane.
    c : float64
        Coefficient of the equation of the plane.
    d : float64
        Coefficient of the equation of the plane.

    Examples
    --------
    >>> N1 = [2.298354000, 5.161785000, 7.971898000]
    >>> N2 = [1.885657000, 4.804777000, 6.183726000]
    >>> N3 = [1.747515000, 6.960963000, 7.932784000]
    >>> a, b, c, d = find_eq_of_plane(N1, N2, N3)
    >>> a
    -3.231203733528
    >>> b
    -0.9688526458499996
    >>> c
    0.9391692927779998
    >>> d
    -4.940497273569501

    """
    x = np.asarray(x, dtype=np.float64)
    y = np.asarray(y, dtype=np.float64)
    z = np.asarray(z, dtype=np.float64)

    xz = z - x
    xy = y - x

    cross_vector = np.cross(xz, xy)
    a, b, c = cross_vector

    d = np.dot(cross_vector, z)

    return a, b, c, d


def find_fit_plane(coord):
    """
    Find best fit plane to the given data points (atoms).

    scipy.optimize.minimize is used to find the least-square plane.

    Parameters
    ----------
    coord : array_like
        Coordinates of selected atom chunk.

    Returns
    -------
    xx : float
        Coefficient of the surface.
    yy : float
        Coefficient of the surface.
    z : float
        Coefficient of the surface.
    abcd : tuple
        Coefficient of the equation of the plane.

    Examples
    --------
    >>> points = [(1.1, 2.1, 8.1),
                  (3.2, 4.2, 8.0),
                  (5.3, 1.3, 8.2),
                  (3.4, 2.4, 8.3),
                  (1.5, 4.5, 8.0),
                  (5.5, 6.7, 4.5)
                  ]
    >>> # To plot the plane, run following commands:
    >>> import matplotlib.pyplot as plt
    >>> # map coordinates for scattering plot
    >>> xs, ys, zs = zip(*points)
    >>> plt.scatter(xs, ys, zs)
    >>> plt.show()

    """
    def plane(x, y, params):
        a = params[0]
        b = params[1]
        c = params[2]
        z = a * x + b * y + c
        return z

    def error(params, points):
        result = 0
        for (x, y, z) in points:
            plane_z = plane(x, y, params)
            diff = abs(plane_z - z)
            result += diff ** 2
        return result

    def cross(a, b):
        return [a[1] * b[2] - a[2] * b[1],
                a[2] * b[0] - a[0] * b[2],
                a[0] * b[1] - a[1] * b[0]]

    points = coord

    fun = functools.partial(error, points=points)
    params0 = [0, 0, 0]
    res = scipy.optimize.minimize(fun, params0)

    a = res.x[0]
    b = res.x[1]
    c = res.x[2]

    point = np.array([0.0, 0.0, c])
    normal = np.array(cross([1, 0, a], [0, 1, b]))
    d = -point.dot(normal)
    xx, yy = np.meshgrid([-5, 10], [-5, 10])
    z = (-normal[0] * xx - normal[1] * yy - d) * 1. / normal[2]

    abcd = (a, b, c, d)

    return xx, yy, z, abcd




#!/usr/bin/env python

# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os

import octadist.main

version = octadist.__version__
release = octadist.__release__


def run_gui():
    """
    OctaDist graphical user interface (GUI).

    """
    print(">>> Program started")
    print(f">>> OctaDist {version} {release}")

    app = octadist.main.OctaDist()
    app.start_app()

    # Delete icon after closing app
    if app.octadist_icon is not None:
        os.remove(app.octadist_icon)

    print(">>> Program terminated")


if __name__ == '__main__':
    run_gui()




# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from math import sqrt, pow, degrees, acos


def angle_sign(v1, v2, direct):
    """
    Compute angle between two vectors with sign and return value in degree.

    Parameters
    ----------
    v1 : array_like
        Vector in 3D space.
    v2 : array_like
        Vector in 3D space.
    direct : array
        Vector that refers to orientation of the plane.

    Returns
    -------
    angle : float64
        Angle between two vectors in degree unit with sign.

    See Also
    --------
    calc.calc_theta

    Examples
    --------
    >>> vector1 = [1.21859514, -0.92569245, -0.51717955]
    >>> vector2 = [1.02186387,  0.57480095, -0.95220433]
    >>> direction = [1.29280503, 0.69301873, 1.80572438]
    >>> angle_sign(vector1, vector2, direction)
    60.38697927455357

    """
    v1 = np.asarray(v1, dtype=np.float64)
    v2 = np.asarray(v2, dtype=np.float64)

    v1 = v1 / np.linalg.norm(v1)
    v2 = v2 / np.linalg.norm(v2)

    angle = np.degrees(np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0)))

    matrix = np.array([v1, v2, direct], dtype=np.float64)
    det = np.float64(np.linalg.det(matrix))

    if det < 0:
        angle = angle * -1

    return angle


def angle_btw_vectors(v1, v2):
    """
    Compute angle between two vectors and return value in degree.

    Parameters
    ----------
    v1 : array_like
        Vector in 3D space.
    v2 : array_like
        Vector in 3D space.

    Returns
    -------
    angle : float64
        Angle between two vectors in degree unit.

    Examples
    --------
    >>> vector1 = [-0.412697, -0.357008, -1.788172]
    >>> vector2 = [-0.550839,  1.799178, -0.039114]
    >>> angle_btw_vectors(vector1, vector2)
    95.62773246517462

    """
    v1 = np.asarray(v1, dtype=np.float64)
    v2 = np.asarray(v2, dtype=np.float64)

    v1 = v1 / np.linalg.norm(v1)
    v2 = v2 / np.linalg.norm(v2)

    angle = np.degrees(np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0)), dtype=np.float64)

    return angle


def angle_btw_planes(a1, b1, c1, a2, b2, c2):
    """
    Find the angle between 2 planes in 3D and return value in degree.

    | General equation of plane:
    |
    | a*X + b*Y + c*Z + d = 0

    Parameters
    ----------
    a1, b1, c1 : float
        Coefficient of the equation of plane 1.
    a2, b2, c2 : float
        Coefficient of the equation of plane 2.

    Returns
    -------
    angle : float64
        Angle between 2 planes in degree unit.

    Examples
    --------
    >>> # Plane 1
    >>> a1 = -3.231203733528
    >>> b1 = -0.9688526458499996
    >>> c1 = 0.9391692927779998
    >>> # Plane 2
    >>> a2 = 1.3904813057000005
    >>> b2 = 3.928502357473003
    >>> c2 = -4.924114034864001
    >>> angle_btw_planes(a1, b1, c1, a2, b2, c2)
    124.89920902358416

    """
    d = (a1 * a2 + b1 * b2 + c1 * c2)
    e1 = sqrt(a1 * a1 + b1 * b1 + c1 * c1)
    e2 = sqrt(a2 * a2 + b2 * b2 + c2 * c2)
    d = d / (e1 * e2)

    angle = np.float64(degrees(acos(d)))

    return angle


def triangle_area(a, b, c):
    """
    Calculate the area of the triangle using the cross product:

    | Area = abs(ab X ac)/2
    |
    | where vector ab = b - a and vector ac = c - a.

    Parameters
    ----------
    a : array_like
        3D Coordinate of point.
    b : array_like
        3D Coordinate of point.
    c : array_like
        3D Coordinate of point.

    Returns
    -------
    area : float64
        The triangle area.

    Examples
    --------
    >>> # Three vertices
    >>> a = [2.298354000, 5.161785000, 7.971898000]
    >>> b = [1.885657000, 4.804777000, 6.183726000]
    >>> c = [1.747515000, 6.960963000, 7.932784000]
    >>> triangle_area(a, b, c)
    1.7508135235821773

    """
    a = np.asarray(a, dtype=np.float64)
    b = np.asarray(b, dtype=np.float64)
    c = np.asarray(c, dtype=np.float64)

    ab = b - a
    ac = c - a
    value = (pow(np.dot(ab[1], ac[2]) - np.dot(ab[2], ac[1]), 2) +
             pow(np.dot(ab[2], ac[0]) - np.dot(ab[0], ac[2]), 2) +
             pow(np.dot(ab[0], ac[1]) - np.dot(ab[1], ac[0]), 2)
             )

    area = np.float64(sqrt(value) / 2)

    return area



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from octadist.src import elements, plane, projection, util


class DrawComplex:
    """
    Display 3D structure of octahedral complex with label for each atoms.

    Parameters
    ----------
    atom : list, None
        Atomic symbols of octahedral structure.
    coord : list, array, tuple, bool, None
        Atomic coordinates of octahedral structure.
    cutoff_global : int or float
        Global cutoff for screening bonds.
        Default value is 2.0.
    cutoff_hydrogen : int or float
        Cutoff for screening hydrogen bonds.
        Default value is 1.2.

    Examples
    --------
    >>> atom = ['Fe', 'N', 'N', 'N', 'O', 'O', 'O']
    >>> coord = [[2.298354000, 5.161785000, 7.971898000],
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]
    >>> test = DrawComplex(atom=atom, coord=coord)
    >>> test.add_atom()
    >>> test.add_bond()
    >>> test.add_legend()
    >>> test.show_plot()

    """
    def __init__(self, atom=None, coord=None, cutoff_global=2.0, cutoff_hydrogen=1.2):
        self.atom = atom
        self.coord = coord
        self.cutoff_global = cutoff_global
        self.cutoff_hydrogen = cutoff_hydrogen

        if self.atom is None:
            raise TypeError("atom is not specified")
        if self.coord is None:
            raise TypeError("coord is not specified")

        self.title_name = 'Display Complex'
        self.title_size = '12'
        self.label_size = '10'
        self.show_title = True
        self.show_axis = True
        self.show_grid = True

        self.atoms_pair = []
        self.bond_list = None

        self.start_plot()

    def start_plot(self):
        """
        Introduce figure to plot.

        """
        self.fig = plt.figure()
        self.ax = Axes3D(self.fig)

        self.ax.set_title('Full complex', fontsize="12")
        # ax = fig.add_subplot(111, projection='3d')

    def add_atom(self):
        """
        Add all atoms to show in figure.

        """
        for i in range(len(self.coord)):
            # Determine atomic number
            n = elements.check_atom(self.atom[i])
            self.ax.scatter(self.coord[i][0],
                            self.coord[i][1],
                            self.coord[i][2],
                            marker='o', linewidths=0.5, edgecolors='black',
                            color=elements.check_color(n), label=f"{self.atom[i]}",
                            s=elements.check_radii(n) * 300)

    def add_symbol(self):
        """
        Add symbol of atoms to show in figure.

        """
        for j in range(len(self.atom)):
            self.ax.text(self.coord[j][0] + 0.1,
                         self.coord[j][1] + 0.1,
                         self.coord[j][2] + 0.1,
                         f"{self.atom[j]},{j}", fontsize=9)

    def add_bond(self):
        """
        Calculate bond distance, screen bond, and add them to show in figure.

        See Also
        --------
        octadist.src.util.find_bonds : Find atomic bonds.

        """
        _, self.bond_list = util.find_bonds(self.atom, self.coord,
                                            self.cutoff_global, self.cutoff_hydrogen)

        for i in range(len(self.bond_list)):
            get_atoms = self.bond_list[i]
            x, y, z = zip(*get_atoms)
            atoms = list(zip(x, y, z))
            self.atoms_pair.append(atoms)

        for i in range(len(self.atoms_pair)):
            merge = list(zip(self.atoms_pair[i][0], self.atoms_pair[i][1]))
            x, y, z = merge
            self.ax.plot(x, y, z, 'k-', color="black", linewidth=2)

    def add_face(self, coord):
        """
        Find the faces of octahedral structure and add those faces to show in figure.

        See Also
        --------
        octadist.src.util.find_faces_octa

        """
        _, c_ref, _, _ = util.find_faces_octa(coord)

        # Added faces
        color_list = ["red", "blue", "green", "yellow",
                      "violet", "cyan", "brown", "grey"]

        for i in range(8):
            # Create array of vertices for 8 faces
            get_vertices = c_ref[i].tolist()
            x, y, z = zip(*get_vertices)
            vertices = [list(zip(x, y, z))]
            self.ax.add_collection3d(Poly3DCollection(vertices, alpha=0.5, color=color_list[i]))

    def add_legend(self):
        """
        Add atoms legend to show in figure.

        References
        ----------
        1. Remove duplicate labels in legend.
            Ref: https://stackoverflow.com/a/26550501/6596684.

        2. Fix size of point in legend.
            Ref: https://stackoverflow.com/a/24707567/6596684.

        """
        # remove duplicate labels
        handles, labels = self.ax.get_legend_handles_labels()
        handle_list, label_list = [], []

        for handle, label in zip(handles, labels):
            if label not in label_list:
                handle_list.append(handle)
                label_list.append(label)
        leg = plt.legend(handle_list, label_list,
                         loc="lower left", scatterpoints=1, fontsize=12)

        # fix size of point in legend
        for i in range(len(leg.legendHandles)):
            leg.legendHandles[i]._sizes = [90]

    def config_plot(self, show_title=True, show_axis=True, show_grid=True, **kwargs):
        """
        Setting configuration for figure.

        Parameters
        ----------
        show_title : bool
            If True, show title of figure.
            If False, not show title of figure.
        show_axis : bool
            If True, show axis of figure.
            If False, not show axis of figure.
        show_grid : bool
            If True, show grid of figure.
            If False, not show grid of figure.
        kwargs : dict (optional)
            title_name : title name of figure.
            title_size : text size of title.
            label_size : text size of axis labels.

        """
        title_name_user = kwargs.get('title_name')
        self.title_size = kwargs.get('title_size')
        self.label_size = kwargs.get('label_size')
        self.show_title = show_title
        self.show_axis = show_axis
        self.show_grid = show_grid

        if title_name_user is not None:
            self.ax.set_title(title_name_user)

        if self.title_size is not None:
            if title_name_user is None:
                title_name_user = self.title_name
            self.ax.set_title(title_name_user, fontsize=self.title_size)

        if self.label_size is not None:
            self.ax.set_xlabel(r'X', fontsize=self.label_size)
            self.ax.set_ylabel(r'Y', fontsize=self.label_size)
            self.ax.set_zlabel(r'Z', fontsize=self.label_size)

        if not self.show_title:
            self.ax.set_title('')
        if not self.show_axis:
            plt.axis('off')
        if not self.show_grid:
            self.ax.grid(False)

    @staticmethod
    def save_img(save="Complex_saved_by_OctaDist", file="png"):
        """
        Save figure as an image.

        Parameters
        ----------
        save : str
            Name of image file.
            Default value is "Complex_saved_by_OctaDist".
        file : file
            Image type.
            Default value is "png".

        """
        plt.savefig(f"{save}.{file}")

    @staticmethod
    def show_plot():
        """
        Show plot.

        """
        plt.show()


class DrawProjection:
    """
    Display the selected 4 faces of octahedral complex.

    Parameters
    ----------
    atom : list
        Atomic symbols of octahedral structure.
    coord : list, array, tuple
        Atomic coordinates of octahedral structure.

    Examples
    --------
    >>> atom = ['Fe', 'N', 'N', 'N', 'O', 'O', 'O']
    >>> coord = [[2.298354000, 5.161785000, 7.971898000],
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]
    >>> test = DrawProjection(atom=atom, coord=coord)
    >>> test.add_atom()
    >>> test.add_symbol()
    >>> test.add_plane()
    >>> test.show_plot()

    """
    def __init__(self, **kwargs):
        self.atom = kwargs.get('atom')
        self.coord = kwargs.get('coord')

        if self.atom is None:
            raise TypeError("atom is not specified")
        if self.coord is None:
            raise TypeError("coord is not specified")

        self.sub_plot = []

        self.start_plot()
        self.shift_plot()

    def start_plot(self):
        """
        Introduce figure to plot.

        """
        self.fig = plt.figure()
        self.st = self.fig.suptitle("4 pairs of opposite planes", fontsize="x-large")

        for i in range(4):
            ax = self.fig.add_subplot(2, 2, int(i + 1), projection='3d')
            ax.set_title(f"Pair {i + 1}")
            self.sub_plot.append(ax)

    def shift_plot(self):
        """
        Shift subplots down. Default value is 0.25.

        """
        self.fig.subplots_adjust(top=0.25)
        self.st.set_y(1.0)

    def add_atom(self):
        """
        Add all atoms to show in figure.

        """
        for i in range(4):
            ax = self.sub_plot[i]
            # Metal
            ax.scatter(self.coord[0][0],
                       self.coord[0][1],
                       self.coord[0][2],
                       color='yellow', marker='o', s=100, linewidths=1,
                       edgecolors='black', label="Metal center")

            # Ligand
            for j in range(1, 7):
                ax.scatter(self.coord[j][0],
                           self.coord[j][1],
                           self.coord[j][2],
                           color='red', marker='o', s=50, linewidths=1,
                           edgecolors='black', label="Ligand atoms")

    def add_symbol(self):
        """
        Add all atoms to show in figure.

        """
        for i in range(4):
            ax = self.sub_plot[i]
            # Metal
            ax.text(self.coord[0][0] + 0.1,
                    self.coord[0][1] + 0.1,
                    self.coord[0][2] + 0.1,
                    self.atom[0], fontsize=9)

            # Ligand
            for j in range(1, 7):
                ax.text(self.coord[j][0] + 0.1,
                        self.coord[j][1] + 0.1,
                        self.coord[j][2] + 0.1,
                        f"{self.atom[j]},{j}", fontsize=9)

    def add_plane(self):
        """
        Add the projection planes to show in figure.

        See Also
        --------
        octadist.src.util.find_faces_octa

        """
        _, c_ref, _, c_oppo = util.find_faces_octa(self.coord)

        color_1 = ["red", "blue", "orange", "magenta"]
        color_2 = ["green", "yellow", "cyan", "brown"]

        for i in range(4):
            ax = self.sub_plot[i]

            # reference face
            get_vertices = c_ref[i].tolist()
            x, y, z = zip(*get_vertices)
            vertices_ref = [list(zip(x, y, z))]

            # opposite face
            x, y, z = zip(*c_oppo[i])
            vertices_oppo = [list(zip(x, y, z))]

            ax.add_collection3d(Poly3DCollection(vertices_ref, alpha=0.5, color=color_1[i]))
            ax.add_collection3d(Poly3DCollection(vertices_oppo, alpha=0.5, color=color_2[i]))

    @staticmethod
    def save_img(save="Complex_saved_by_OctaDist", file="png"):
        """
        Save figure as an image.

        Parameters
        ----------
        save : str
            Name of image file.
            Default value is "Complex_saved_by_OctaDist".
        file : file
            Image type.
            Default value is "png".

        """
        plt.savefig(f"{save}.{file}")

    @staticmethod
    def show_plot():
        """
        Show plot.

        """
        plt.tight_layout()
        plt.show()


class DrawTwistingPlane:
    """
    Display twisting triangular faces and vector projection.

    Parameters
    ----------
    atom : list
        Atomic symbols of octahedral structure.
    coord : list, array, tuple
        Atomic coordinates of octahedral structure.

    Examples
    --------
    >>> atom = ['Fe', 'N', 'N', 'N', 'O', 'O', 'O']
    >>> coord = [[2.298354000, 5.161785000, 7.971898000],
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]
    >>> test = DrawTwistingPlane(atom=atom, coord=coord)
    >>> test.add_plane()
    >>> test.add_symbol()
    >>> test.add_bond()
    >>> test.show_plot()

    """
    def __init__(self, **kwargs):
        self.atom = kwargs.get('atom')
        self.coord = kwargs.get('coord')

        if self.atom is None:
            raise TypeError("atom is not specified")
        if self.coord is None:
            raise TypeError("coord is not specified")

        _, self.c_ref, _, self.c_oppo = util.find_faces_octa(self.coord)

        self.all_ax = []
        self.all_m = []
        self.all_proj_ligs = []

        self.start_plot()
        self.shift_plot()
        self.create_subplots()

    def start_plot(self):
        """
        Introduce figure to plot.

        """
        self.fig = plt.figure()
        self.st = self.fig.suptitle("Projected twisting triangular faces", fontsize="x-large")

    def shift_plot(self):
        """
        Shift subplots down. Default value is 0.25.

        """
        self.fig.subplots_adjust(top=0.25)
        self.st.set_y(1.0)

    def create_subplots(self):
        """
        Create subplots.

        """
        for i in range(4):
            ax = self.fig.add_subplot(2, 2, int(i + 1), projection='3d')
            ax.set_title(f"Projection plane {i + 1}", fontsize='10')

            self.all_ax.append(ax)

    def add_plane(self):
        """
        Add the projection planes to show in figure.

        See Also
        --------
        octadist.src.plane.find_eq_of_plane :
            Find the equation of the plane.
        octadist.src.projection.project_atom_onto_plane :
            Orthogonal projection of point onto the plane.

        """
        for i in range(4):
            a, b, c, d = plane.find_eq_of_plane(self.c_ref[i][0],
                                                self.c_ref[i][1],
                                                self.c_ref[i][2])
            m = projection.project_atom_onto_plane(self.coord[0], a, b, c, d)
            self.all_m.append(m)

            ax = self.all_ax[i]

            # Projected metal center atom
            ax.scatter(m[0],
                       m[1],
                       m[2],
                       color='orange', s=100, marker='o', linewidths=1,
                       edgecolors='black', label="Metal center")

            # Reference atoms
            all_proj_lig = []
            for j in range(3):
                ax.scatter(self.c_ref[i][j][0],
                           self.c_ref[i][j][1],
                           self.c_ref[i][j][2],
                           color='red', s=50, marker='o', linewidths=1,
                           edgecolors='black', label="Reference atom")

                # Project ligand atom onto the reference face
                proj_lig = projection.project_atom_onto_plane(self.c_oppo[i][j], a, b, c, d)
                all_proj_lig.append(proj_lig)

            # Projected opposite atoms
                ax.scatter(proj_lig[0],
                           proj_lig[1],
                           proj_lig[2],
                           color='blue', s=50, marker='o', linewidths=1,
                           edgecolors='black', label="Projected ligand atom")

            self.all_proj_ligs.append(all_proj_lig)

            # Draw plane
            get_vertices = self.c_ref[i].tolist()
            x, y, z = zip(*get_vertices)
            vertices = [list(zip(x, y, z))]

            x, y, z = zip(*self.all_proj_ligs[i])
            projected_oppo_vertices_list = [list(zip(x, y, z))]
            ax.add_collection3d(Poly3DCollection(vertices,
                                                 alpha=0.5,
                                                 color="yellow"))
            ax.add_collection3d(Poly3DCollection(projected_oppo_vertices_list,
                                                 alpha=0.5,
                                                 color="blue"))

    def add_symbol(self):
        """
        Add all atoms to show in figure.

        """
        for i in range(4):
            ax = self.all_ax[i]
            ax.text(self.all_m[i][0] + 0.1,
                    self.all_m[i][1] + 0.1,
                    self.all_m[i][2] + 0.1,
                    f"{self.atom[0]}'", fontsize=9)

            for j in range(3):
                ax.text(self.c_ref[i][j][0] + 0.1,
                        self.c_ref[i][j][1] + 0.1,
                        self.c_ref[i][j][2] + 0.1,
                        f"{j + 1}", fontsize=9)

                ax.text(self.all_proj_ligs[i][j][0] + 0.1,
                        self.all_proj_ligs[i][j][1] + 0.1,
                        self.all_proj_ligs[i][j][2] + 0.1,
                        f"{j + 1}'", fontsize=9)

    def add_bond(self):
        """
        Calculate bond distance, screen bond, and add them to show in figure.

        """
        for i in range(4):
            for j in range(3):
                merge = list(zip(self.all_m[i].tolist(), self.c_ref[i][j].tolist()))
                x, y, z = merge
                self.all_ax[i].plot(x, y, z, 'k-', color="black")

                merge = list(zip(self.all_m[i].tolist(), self.all_proj_ligs[i][j].tolist()))
                x, y, z = merge
                self.all_ax[i].plot(x, y, z, 'k->', color="black")

    @staticmethod
    def save_img(save="Complex_saved_by_OctaDist", file="png"):
        """
        Save figure as an image.

        Parameters
        ----------
        save : str
            Name of image file.
            Default value is "Complex_saved_by_OctaDist".
        file : file
            Image type.
            Default value is "png".

        """
        plt.savefig(f"{save}.{file}")

    @staticmethod
    def show_plot():
        """
        Show plot.

        """
        # plt.legend(bbox_to_anchor=(1.05, 1), loc=2)
        plt.tight_layout()
        plt.show()



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from tkinter.messagebox import showinfo, showerror, showwarning


def err_no_file():
    """
    Show this error when having no any input files is loaded into program.

    """
    showerror("Error", "No input file. At least one input file must be loaded.")


def err_invalid_ftype():
    """
    Show this error popup when file type is not supported by the program.

    """
    showerror("Error", f"Invalid input file format.\n\n"
              "The complex must have at least 1 metal atom and 6 ligand atoms.")


def err_no_coord(i):
    """
    Show this error popup when the program cannot read the atomic coordinates of complex
    inside the file or cannot extract the coordinates from the complex.

    This will happen only if the input has no the proper format of atomic coordinates.

    Parameters
    ----------
    i : int
        Number of file.

    """
    showerror("Error", f"File no. {i} has no atomic coordinates.\n\n"
              "Please make sure the input file format is correct.")


def err_less_ligands(i):
    """
    Show this error popup when the complex has ligand atoms less that six.

    Parameters
    ----------
    i : int
        Number of file.

    """
    showerror("Error", f"File no. {i} has ligand atoms less than six atoms.\n\n"
              "Please check if some ligand atoms are so far away from metal atom,"
              "or check if the metal-ligand bond cutoff is set properly.")


def err_no_metal():
    """
    Show this error popup when the complex has no transition metal atom.

    """
    showerror("Error", "Cannot proceed calculation. Your current structure has no transition metal.\n\n"
                       "The complex must have at least one transition metal.")


def err_no_calc():
    """
    Show this error popup when the user requests function that the results are required,
    but the results have not been computed yet.

    """
    showerror("Error", "No results. Click \"Compute\" to calculate octahedral distortion parameters.")


def err_only_2_files():
    """
    Show this error popup when having not uploaded two complexes
    for using the RMSD function.

    """
    showerror("Error", "You must load only two complexes for computing RMSD.")


def err_not_equal_atom():
    """
    Show this error popup when the total number of atoms of two complexes are not equal.

    """
    showerror("Error", "The number of atoms in structure 1 and structure 2 are not the same.")


def err_atom_not_match(line):
    """
    show this error popup when atomic symbol of two similar complexes does not match.

    Parameters
    ----------
    line : int
        The line number that atomic symbol does not match.

    """
    showerror("Error", f"Atomic symbol not match at line {line}.")


def err_many_files():
    """
    Show this error popup when user has loaded too many files.

    """
    showerror("Error", "You must load only one input file.")


def err_wrong_format():
    """
    Show this error popup when user has loaded the file that is not supported by OctaDist.

    """
    showerror("Error", "Could not read file.\n\n"
                       "Input file format is not supported by the current version of OctaDist.")


def err_no_editor():
    """
    Show this error popup if text editor path is empty.

    """
    showerror("Error", "No text editor.\n\n"
                       "Text editor path is empty. Go to setting and browse a new text editor.")


def err_cannot_update():
    """
    Show this error popup when the program cannot detect the operating system of
    the machine that user is using.

    """
    showerror("Error", "Cannot download an installer of a new version.\n\n"
                       "Please contact OctaDist development team for further help.")


def info_save_results(file):
    """
    Show info popup:

    show this info popup when output file has been saved successfully.

    Parameters
    ----------
    file : str
        Absolute or full path of saved output file.

    """
    showinfo("Info", f"Data has been saved to {file}")


def info_new_update():
    """
    Show this info popup when new version is available for update.

    """
    showinfo("Info", "New updates available!")


def info_using_dev():
    """
    Show this info popup if user is using a development build version.

    """
    showinfo("Info", "You are using a development build version!")


def info_no_update():
    """
    Show this info popup if program is the latest version.

    """
    showinfo("Info", "You already have the latest version.")


def warn_no_metal(i):
    """
    Show this warning popup if no transition metal was found.

    Parameters
    ----------
    i : int
        Number of file.

    """
    showwarning("Warning", f"File no. {i} has no transition metal in your complex.")


def warn_not_octa():
    """
    Show this warning popup if the complex is non-octahedral structure.

    """
    showwarning("Warning", "Non-octahedral complex detected.")



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import argparse
import os
import sys

import octadist
from .octadist_gui import run_gui
from octadist.src.molecule import (
    is_xyz, get_coord_xyz, extract_octa
)


def check_file(file):
    """
    Check if input file is exist or not.

    Parameters
    ----------
    file : str
        Input file name.

    Returns
    -------
    file : str
        Input file name.

    """
    exist = os.path.isfile(file)

    if exist:
        return file
    else:
        print(f"File not found: {file}")
        sys.exit(1)


def find_coord(file):
    """
    Find atomic symbols and atomic coordinates of structure.

    Parameters
    ----------
    file : str
        Input file name.

    Returns
    -------
    atom : list
        Atomic symbols.
    coord : list
        Atomic coordinates.

    """
    if file.endswith(".xyz"):
        if is_xyz(file):
            atom, coord = get_coord_xyz(file)
        else:
            print(f"File type of input file is not supported: {file}")
            sys.exit(1)
    else:
        print(f"File type of input file is not supported: {file}")
        sys.exit(1)

    atom = list(filter(None, atom))

    return atom, coord


def find_octa(atom, coord):
    """
    Find atomic symbols and atomic coordinates of structure.

    Parameters
    ----------
    atom : list
        Atomic symbols of structure.
    coord : list
        Atomic coordinates of structure.

    Returns
    -------
    atom : list
        Atomic symbols of octahedral structure.
    coord : ndarray
        Atomic coordinates of octahedral structure.

    """
    atom, coord = extract_octa(atom, coord)

    return atom, coord


def calc_param(coord):
    """
    Calculate octahedral distortion parameters.

    Parameters
    ----------
    coord : array_like
        Atomic coordinates of octahedral structure.

    Returns
    -------
    computed : dict
        Computed parameters.

    """
    dist = octadist.CalcDistortion(coord)
    zeta = dist.zeta  # Zeta
    delta = dist.delta  # Delta
    sigma = dist.sigma  # Sigma
    theta = dist.theta  # Theta

    computed = {'zeta': zeta,
                'delta': delta,
                'sigma': sigma,
                'theta': theta
                }

    return computed


def run_cli():
    """
    OctaDist command-line interface (CLI).
    This function has been implemented by entry points function
    in setuptools package.

    """
    description = """\
Octahedral Distortion Calculator: 
A tool for computing octahedral distortion parameters in coordination complex.
For more details, please visit https://github.com/OctaDist/OctaDist.
"""

    epilog = f"Rangsiman Ketkaew\tUpdated on {octadist.__release__}\tE-mail: {octadist.__email__}"

    parser = argparse.ArgumentParser(
        prog="octadist_cli",
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog
    )

    # info
    parser.add_argument('-v', '--version',
                        action='version',
                        version=octadist.__version__
                        )
    parser.add_argument('-a', '--about',
                        action='store_true',
                        help='show program info'
                        )
    parser.add_argument('-c', '--cite',
                        action='store_true',
                        help='show how to cite OctaDist'
                        )
    parser.add_argument('-g', '--gui',
                        action='store_true',
                        help='launch OctaDist GUI (this option is the same as \'octadist\' command'
                        )

    # input/output
    parser.add_argument('-i', '--inp',
                        action='store',
                        type=str,
                        metavar='INPUT',
                        help='input structure in .xyz format'
                        )
    parser.add_argument('-o', '--out',
                        action='store_true',
                        help='show formatted output summary'
                        )
    parser.add_argument('-s', '--save',
                        action='store',
                        type=str,
                        metavar='OUTPUT',
                        help='save formatted output to text file, '
                             'please specify name of OUTPUT file without \'.txt\' extension'
                        )

    # result
    parser.add_argument('--par',
                        type=str,
                        nargs='+',
                        choices=['zeta', 'delta', 'sigma', 'theta'],
                        metavar='PARAMETER',
                        help='select which the parameter (zeta, delta, sigma, theta) to show'
                        )
    parser.add_argument('--show',
                        type=str,
                        nargs='+',
                        choices=['atom', 'coord'],
                        metavar='MOL',
                        help='show atomic symbol (atom) and atomic coordinate (coord) of '
                             'octahedral structure'
                        )

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    if args.about:
        print("\nOctaDist Program info")
        print("=====================")
        print(f"- Name\t\t=\t{octadist.__title__}")
        print(f"- Author\t=\t{octadist.__author__}")
        print(f"- Version\t=\t{octadist.__version__}")
        print(f"- Revision\t=\t{octadist.__revision__}")
        print(f"- Release\t=\t{octadist.__release__}")
        print(f"- Description\t=\t{octadist.__description__}")
        print(f"- E-mail\t=\t{octadist.__email__}")
        print(f"- Document\t=\t{octadist.__doc__}")
        print(f"- Website\t=\t{octadist.__website__}")
        sys.exit(1)

    if args.cite:
        print("\nHow to cite OctaDist")
        print("=====================")
        print("Please cite this project when you use OctaDist for scientific publication.\n")
        print("  OctaDist: A tool for calculating distortion parameters in coordination complexes.")
        print("  https://octadist.github.io\n")
        sys.exit(1)

    # in case GUI is requested
    if args.gui:
        run_gui()
        sys.exit(1)

    atom_coord = {}
    computed = {}
    token = False

    # find coordinates of structure
    if args.inp:
        # check if file is correct
        file = check_file(args.inp)

        atom, coord = find_coord(file)
        atom, coord = find_octa(atom, coord)

        atom_coord = {'atom': atom,
                      'coord': coord
                      }

        computed = calc_param(coord)

        token = True
    else:
        print("No input file specified")

    # get only basename of file from path
    basename = os.path.basename(args.inp)

    # print computed parameters
    if not args.show:
        if not args.out:
            if not args.par and token:
                for key in ['zeta', 'delta', 'sigma', 'theta']:
                    print(computed[key])
            else:
                for key in args.par:
                    print(computed[key])
        else:
            print("Octahedral distortion parameters")
            print("--------------------------------")
            print(f"File: {basename}")
            print(f"Zeta   = {computed['zeta']:12.8f}")
            print(f"Delta  = {computed['delta']:12.8f}")
            print(f"Sigma  = {computed['sigma']:12.8f}")
            print(f"Theta  = {computed['theta']:12.8f}")

    # print atom and coord
    if args.show and token:
        for key in args.show:
            print(atom_coord[key])

    # save result
    if args.save and token:
        with open(args.save + '.txt', 'w') as f:
            f.write("Octahedral distortion parameters\n")
            f.write("--------------------------------\n")
            f.write(f"File: {basename}\n")
            f.write(f"Zeta   = {computed['zeta']:12.8f}\n")
            f.write(f"Delta  = {computed['delta']:12.8f}\n")
            f.write(f"Sigma  = {computed['sigma']:12.8f}\n")
            f.write(f"Theta  = {computed['theta']:12.8f}\n")
            f.write(f"\nComputed by OctaDist {octadist.__version__}\n")
            f.close()
        print(f"\nOutput file has been saved to {os.path.realpath(f.name)}")


if __name__ == '__main__':
    run_cli()



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from scipy.spatial import distance

from octadist.src import linear, plane, projection


class CalcDistortion:
    """
    Calculate distortion parameters:
    - Bond distance
    - Mean bond distance
    - Bond angle around metal center atom
    - Delta parameter
    - Sigma parameter
    - Minimum, maximum, and mean Theta parameters

    Parameters
    ----------
    coord : array_like
        Atomic coordinates of octahedral structure.

    Examples
    --------
    >>> coord = [[2.298354000, 5.161785000, 7.971898000],  # <- Metal atom
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]
    >>> test = CalcDistortion(coord)
    >>> test.sigma
    47.926528379270124

    """

    def __init__(self, coord):
        self.coord = coord

        if type(self.coord) == np.ndarray:
            pass
        else:
            self.coord = np.asarray(self.coord, dtype=np.float64)

        self.bond_dist = []
        self.d_mean = 0
        self.diff_dist = []
        self.zeta = 0
        self.delta = 0
        self.cis_angle = 0
        self.trans_angle = 0
        self.sigma = 0
        self.eight_theta = []
        self.theta = 0
        self.theta_min = 0
        self.theta_max = 0
        self.eq_of_plane = []
        self.non_octa = False

        self.calc_d_bond()
        self.calc_d_mean()
        self.calc_bond_angle()
        self.calc_zeta()
        self.calc_delta()
        self.calc_sigma()
        self.calc_theta()
        self.calc_theta_min()
        self.calc_theta_max()

    def calc_d_bond(self):
        """
        Calculate metal-ligand bond distance and return value in Angstrom.

        See Also
        --------
        calc_d_mean : Calculate mean metal-ligand bond length.

        """
        self.bond_dist = [distance.euclidean(self.coord[0], self.coord[i]) for i in range(1, 7)]
        self.bond_dist = np.asarray(self.bond_dist, dtype=np.float64)

    def calc_d_mean(self):
        """
        Calculate mean distance parameter and return value in Angstrom.

        See Also
        --------
        calc_d_bond : Calculate metal-ligand bonds length.

        """
        self.d_mean = np.mean(self.bond_dist)

    def calc_bond_angle(self):
        """
        Calculate 12 cis and 3 trans unique angles in octahedral structure.

        See Also
        --------
        calc_sigma : Calculate Sigma parameter.

        """
        all_angle = []
        for i in range(1, 7):
            for j in range(i + 1, 7):
                vec1 = self.coord[i] - self.coord[0]
                vec2 = self.coord[j] - self.coord[0]
                angle = linear.angle_btw_vectors(vec1, vec2)
                all_angle.append(angle)

        # Sort the angle from the lowest to the highest
        sorted_angle = sorted(all_angle)
        self.cis_angle = [sorted_angle[i] for i in range(12)]
        self.trans_angle = [sorted_angle[i] for i in range(12, 15)]

    def calc_zeta(self):
        """
        Calculate Zeta parameter and return value in Angstrom.

        See Also
        --------
        calc_d_bond : Calculate metal-ligand bonds length.
        calc_d_mean : Calculate mean metal-ligand bond length.

        References
        ----------
        .. [1] M. Buron-Le Cointe, J. Hébert, C. Baldé, N. Moisan,
            L. Toupet, P. Guionneau, J. F. Létard, E. Freysz,
            H. Cailleau, and E. Collet. - Intermolecular control of
            thermoswitching and photoswitching phenomena in two
            spin-crossover polymorphs. Phys. Rev. B 85, 064114.

        """
        diff_dist = [abs(self.bond_dist[i] - self.d_mean) for i in range(6)]
        self.diff_dist = np.asarray(diff_dist, dtype=np.float64)

        self.zeta = np.sum(self.diff_dist)

    def calc_delta(self):
        """
        Calculate Delta parameter, also known as Tilting distortion parameter.

        See Also
        --------
        calc_d_bond : Calculate metal-ligand bonds length.
        calc_d_mean : Calculate mean metal-ligand bond length.

        References
        ----------
        .. [2] M. W. Lufaso and P. M. Woodward. - Jahn–Teller distortions,
            cation ordering and octahedral tilting in perovskites.
            Acta Cryst. (2004). B60, 10-20. DOI: 10.1107/S0108768103026661

        """
        delta = sum(pow((self.bond_dist[i] - self.d_mean) / self.d_mean, 2) for i in range(6))
        self.delta = delta / 6

    def calc_sigma(self):
        """
        Calculate Sigma parameter and return value in degree.

        See Also
        --------
        calc_bond_angle : Calculate bond angles between ligand-metal-ligand.

        References
        ----------
        .. [3] James K. McCusker, A. L. Rheingold, D. N. Hendrickson.
            Variable-Temperature Studies of Laser-Initiated 5T2 → 1A1
            Intersystem Crossing in Spin-Crossover Complexes: 
            Empirical Correlations between Activation Parameters
            and Ligand Structure in a Series of Polypyridyl.
            Ferrous Complexes. Inorg. Chem. 1996, 35, 2100.

        """
        self.sigma = sum(abs(90.0 - self.cis_angle[i]) for i in range(12))

    def determine_faces(self):
        """
        Refine the order of ligand atoms in order to find the plane
        for projection.

        Returns
        -------
        coord_metal : nd.array
            Coordinate of metal atom.
        coord_lig : nd.array
            Coordinate of ligand atoms.

        See Also
        --------
        calc_theta :
            Calculate mean Theta parameter

        Examples
        --------
        >>> bef = np.array([[4.0674, 7.2040, 13.6117]
                            [4.3033, 7.3750, 11.7292]
                            [3.8326, 6.9715, 15.4926]
                            [5.8822, 6.4461, 13.4312]
                            [3.3002, 5.3828, 13.6316]
                            [4.8055, 8.9318, 14.2716]
                            [2.3184, 8.0165, 13.1152]
                            ])

        >>> metal, coord = self.determine_faces(bef)
        >>> metal
        [ 4.0674  7.204  13.6117]
        >>> coord_lig
        [[ 4.3033  7.375  11.7292]      # Front face
         [ 4.8055  8.9318 14.2716]      # Front face
         [ 5.8822  6.4461 13.4312]      # Front face
         [ 2.3184  8.0165 13.1152]      # Back face
         [ 3.8326  6.9715 15.4926]      # Back face
         [ 3.3002  5.3828 13.6316]]     # Back face

        """
        # Metal and ligand atoms
        coord_metal = self.coord[0]
        ligands = self.coord[1:]
        coord_lig = np.array([self.coord[i] for i in range(1, 7)])

        # Find vector from metal to ligand atoms
        metal_to_lig = coord_lig - coord_metal

        # Find maximum angle
        max_angle = self.trans_angle[0]

        # Identify which N is in line with ligand 1
        def_change = 6
        for n in range(6):
            test = linear.angle_btw_vectors(metal_to_lig[0], metal_to_lig[n])
            if test > (max_angle - 1):
                def_change = n

        test_max = 0
        new_change = 0
        for n in range(6):
            test = linear.angle_btw_vectors(metal_to_lig[0], metal_to_lig[n])
            if test > test_max:
                test_max = test
                new_change = n

        # Check if the structure is octahedron or not
        if def_change != new_change:
            self.non_octa = True
            def_change = new_change

        # Swap ligand
        # As ligands in 2-dimensions array, we use the following technique
        # to swap two lists of elements with avoiding passing reference value
        ligands[[4, def_change]] = ligands[[def_change, 4]]

        # Update vector from metal to ligand atoms
        metal_to_lig = ligands - coord_metal

        # Identify which N is in line with ligand 2
        for n in range(6):
            test = linear.angle_btw_vectors(metal_to_lig[1], metal_to_lig[n])
            if test > (max_angle - 1):
                def_change = n

        test_max = 0
        for n in range(6):
            test = linear.angle_btw_vectors(metal_to_lig[1], metal_to_lig[n])
            if test > test_max:
                test_max = test
                new_change = n

        if def_change != new_change:
            self.non_octa = True
            def_change = new_change

        # Swap ligand
        ligands[[5, def_change]] = ligands[[def_change, 5]]

        # Update vector from metal to ligand atoms
        metal_to_lig = ligands - coord_metal

        # Identify which N is in line with ligand 3
        for n in range(6):
            test = linear.angle_btw_vectors(metal_to_lig[2], metal_to_lig[n])
            if test > (max_angle - 1):
                def_change = n

        test_max = 0
        for n in range(6):
            test = linear.angle_btw_vectors(metal_to_lig[2], metal_to_lig[n])
            if test > test_max:
                test_max = test
                new_change = n

        if def_change != new_change:
            self.non_octa = True
            def_change = new_change

        # Swap ligand
        ligands[[3, def_change]] = ligands[[def_change, 3]]

        # New atom order
        coord_lig = np.array([ligands[i] for i in range(6)])

        return coord_metal, coord_lig

    def calc_theta(self):
        """
        Calculate Theta parameter and value in degree.

        See Also
        --------
        calc_theta_min :
            Calculate minimum Theta parameter.
        calc_theta_max :
            Calculate maximum Theta parameter.
        octadist.src.linear.angle_btw_vectors :
            Calculate cosine angle between two vectors.
        octadist.src.linear.angle_sign
            Calculate cosine angle between two vectors sensitive to CW/CCW direction.
        octadist.src.plane.find_eq_of_plane :
            Find the equation of the plane.
        octadist.src.projection.project_atom_onto_plane :
            Orthogonal projection of point onto the plane.

        References
        ----------
        .. [4] M. Marchivie, P. Guionneau, J.-F. Létard, D. Chasseau.
            Photo‐induced spin‐transition: the role of the iron(II)
            environment distortion. Acta Crystal-logr. Sect. B Struct.
            Sci. 2005, 61, 25.

        """
        # Get refined atomic coordinates
        coord_metal, coord_lig = self.determine_faces()

        # loop over 8 faces
        for r in range(8):
            a, b, c, d = plane.find_eq_of_plane(coord_lig[0], coord_lig[1], coord_lig[2])
            self.eq_of_plane.append([a, b, c, d])

            # Project metal and other three ligand atom onto the plane
            projected_m = projection.project_atom_onto_plane(coord_metal, a, b, c, d)
            projected_lig4 = projection.project_atom_onto_plane(coord_lig[3], a, b, c, d)
            projected_lig5 = projection.project_atom_onto_plane(coord_lig[4], a, b, c, d)
            projected_lig6 = projection.project_atom_onto_plane(coord_lig[5], a, b, c, d)

            # Find the vectors between atoms that are on the same plane
            # These vectors will be used to calculate Theta afterward.
            vector_theta = np.array([
                coord_lig[0] - projected_m,
                coord_lig[1] - projected_m,
                coord_lig[2] - projected_m,
                projected_lig4 - projected_m,
                projected_lig5 - projected_m,
                projected_lig6 - projected_m
            ])

            # Check if the direction is CW or CCW
            a12 = linear.angle_btw_vectors(vector_theta[0], vector_theta[1])
            a13 = linear.angle_btw_vectors(vector_theta[0], vector_theta[2])

            # If angle of interest is smaller than its neighbor,
            # define it as CW direction, if not, it will be CCW instead.
            if a12 < a13:
                direction = np.cross(vector_theta[0], vector_theta[1])
            else:
                direction = np.cross(vector_theta[2], vector_theta[0])

            # Calculate individual theta angle
            theta1 = linear.angle_sign(vector_theta[0], vector_theta[3], direction)
            theta2 = linear.angle_sign(vector_theta[3], vector_theta[1], direction)
            theta3 = linear.angle_sign(vector_theta[1], vector_theta[4], direction)
            theta4 = linear.angle_sign(vector_theta[4], vector_theta[2], direction)
            theta5 = linear.angle_sign(vector_theta[2], vector_theta[5], direction)
            theta6 = linear.angle_sign(vector_theta[5], vector_theta[0], direction)

            indi_theta = np.array([theta1, theta2, theta3, theta4,  theta5, theta6])

            self.eight_theta.append(sum(abs(indi_theta - 60)))

            # Use deep copy so as to avoid pass by reference
            tmp = coord_lig[1].copy()
            coord_lig[1] = coord_lig[3].copy()
            coord_lig[3] = coord_lig[5].copy()
            coord_lig[5] = coord_lig[2].copy()
            coord_lig[2] = tmp.copy()

            # If 3rd round, permutation face will be switched
            # from N1N2N3
            # to N1N4N2,
            # to N1N6N4,
            # to N1N3N6, and then back to N1N2N3
            if r == 3:
                coord_lig[[0, 4]] = coord_lig[[4, 0]]
                coord_lig[[1, 5]] = coord_lig[[5, 1]]
                coord_lig[[2, 3]] = coord_lig[[3, 2]]

        self.theta = sum(self.eight_theta) / 2

    def calc_theta_min(self):
        """
        Calculate minimum Theta parameter and return value in degree.

        See Also
        --------
        calc_theta :
            Calculate mean Theta parameter

        """
        sorted_theta = sorted(self.eight_theta)
        self.theta_min = sum(sorted_theta[i] for i in range(4))

    def calc_theta_max(self):
        """
        Calculate maximum Theta parameter and return value in degree.

        See Also
        --------
        calc_theta :
            Calculate mean Theta parameter

        """
        sorted_theta = sorted(self.eight_theta)
        self.theta_max = sum(sorted_theta[i] for i in range(4, 8))



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import tkinter as tk
from tkinter import scrolledtext as tkscrolled

import numpy as np
import rmsd

from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from octadist.src import linear, elements, plane, util


class CalcJahnTeller:
    """
    Calculate angular Jahn-Teller distortion parameter.

    Parameters
    ----------
    atom : array_like
        Atomic labels of full complex.
    coord : array_like
        Atomic coordinates of full complex.
    master : None, object
        If None, use tk.Tk().
        If not None, use tk.Toplevel(master).
    cutoff_global : int or float
        Global cutoff for screening bonds.
        Default value is 2.0.
    cutoff_hydrogen : int or float
        Cutoff for screening hydrogen bonds.
        Default value is 1.2.
    icon : str, optional
        If None, use tkinter default icon.
        If not None, use user-defined icon.

    Examples
    --------
    >>> atom = ['Fe', 'N', 'N', 'N', 'O', 'O', 'O']
    >>> coord = [[2.298354000, 5.161785000, 7.971898000],
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]

    >>> test = CalcJahnTeller(atom=atom, coord=coord)
    >>> test.start_app()
    >>> test.find_bond()
    >>> test.show_app()

    References
    ----------
    .. [1] J. M. Holland, J. A. McAllister, C. A. Kilner,
        M. Thornton-Pett, A. J. Bridgeman, M. A. Halcrow.
        Stereochemical effects on the spin-state transition
        shown by salts of [FeL2]2+ [L = 2,6-di(pyrazol-1-yl)pyridine].
        J. Chem. Soc., Dalton Trans., 2002, 548-554.
        DOI: 10.1039/B108468M.

    """

    def __init__(self, atom, coord, cutoff_global=2.0, cutoff_hydrogen=1.2, master=None, icon=None):
        self.atom = atom
        self.coord = coord
        self.cutoff_global = cutoff_global
        self.cutoff_hydrogen = cutoff_hydrogen

        if master is None:
            self.wd = tk.Tk()
        else:
            self.wd = tk.Toplevel(master)

        self.icon = icon

        self.bond_list = []
        self.coord_A = []
        self.coord_B = []

    def start_app(self):
        """
        Start application.

        """
        if self.icon is not None:
            self.wd.wm_iconbitmap(self.icon)
        self.wd.title("Calculate Jahn-Teller distortion parameter")
        self.wd.geometry("630x550")
        self.wd.resizable(0, 0)

        self.lbl = tk.Label(self.wd, text="Group A")
        self.lbl.config(width=12)
        self.lbl.grid(padx="10", pady="5", row=0, column=0, columnspan=2)

        self.lbl = tk.Label(self.wd, text="Group B")
        self.lbl.config(width=12)
        self.lbl.grid(padx="10", pady="5", row=0, column=2, columnspan=2)

        self.box_1 = tkscrolled.ScrolledText(self.wd, height="12", width="40", wrap="word", undo="True")
        self.box_1.grid(padx="5", pady="5", row=1, column=0, columnspan=2)

        self.box_2 = tkscrolled.ScrolledText(self.wd, height="12", width="40", wrap="word", undo="True")
        self.box_2.grid(padx="5", pady="5", row=1, column=2, columnspan=2)

        self.btn = tk.Button(self.wd, text="Select ligand set A", command=lambda: self.pick_atom(group="A"))
        self.btn.config(width=15, relief=tk.RAISED)
        self.btn.grid(padx="10", pady="5", row=2, column=0, columnspan=2)

        self.btn = tk.Button(self.wd, text="Select ligand set B", command=lambda: self.pick_atom(group="B"))
        self.btn.config(width=15, relief=tk.RAISED)
        self.btn.grid(padx="10", pady="5", row=2, column=2, columnspan=2)

        self.btn = tk.Button(self.wd, text="Calculate parameter", command=lambda: self.plot_fit_plane())
        self.btn.config(width=15, relief=tk.RAISED)
        self.btn.grid(padx="10", pady="5", row=3, column=0, columnspan=2)

        self.btn = tk.Button(self.wd, text="Clear all", command=lambda: self.clear_text())
        self.btn.config(width=15, relief=tk.RAISED)
        self.btn.grid(padx="10", pady="5", row=3, column=2, columnspan=2)

        self.lbl = tk.Label(self.wd, text="Supplementary angles between two planes (in degree)")
        self.lbl.grid(pady="10", row=4, columnspan=4)

        self.lbl_angle1 = tk.Label(self.wd, text="Angle 1")
        self.lbl_angle1.grid(pady="5", row=5, column=0)
        self.box_angle1 = tk.Entry(self.wd, width="20", justify='center')
        self.box_angle1.grid(row=5, column=1, sticky=tk.W)

        self.lbl_angle2 = tk.Label(self.wd, text="Angle 2")
        self.lbl_angle2.grid(pady="5", row=6, column=0)
        self.box_angle2 = tk.Entry(self.wd, width="20", justify='center')
        self.box_angle2.grid(row=6, column=1, sticky=tk.W)

        self.lbl = tk.Label(self.wd, text="The equation of the planes")
        self.lbl.grid(pady="10", row=7, columnspan=4)

        self.lbl_eq1 = tk.Label(self.wd, text="Plane A ")
        self.lbl_eq1.grid(pady="5", row=8, column=0)
        self.box_eq1 = tk.Entry(self.wd, width="60", justify='center')
        self.box_eq1.grid(pady="5", row=8, column=1, columnspan=2, sticky=tk.W)

        self.lbl_eq2 = tk.Label(self.wd, text="Plane B ")
        self.lbl_eq2.grid(pady="5", row=9, column=0)
        self.box_eq2 = tk.Entry(self.wd, width="60", justify='center')
        self.box_eq2.grid(pady="5", row=9, column=1, columnspan=2, sticky=tk.W)

    def find_bond(self):
        """
        Find bonds.

        See Also
        --------
        octadist.src.util.find_bonds : Find atomic bonds.

        """
        _, self.bond_list = util.find_bonds(self.atom,
                                            self.coord,
                                            self.cutoff_global,
                                            self.cutoff_hydrogen)

    #################
    # Picking atoms #
    #################

    def pick_atom(self, group):
        """
        On-mouse pick atom and get XYZ coordinate.

        Parameters
        ----------
        group : str
            Group A or B.

        """
        fig = plt.figure()
        # fig = plt.figure(figsize=(5, 4), dpi=100)
        ax = Axes3D(fig)
        # ax = fig.add_subplot(111, projection='3d')

        # Plot all atoms
        for i in range(len(self.coord)):
            # Determine atomic number
            n = elements.check_atom(self.atom[i])
            ax.scatter(self.coord[i][0], self.coord[i][1], self.coord[i][2],
                       marker='o', linewidths=0.5, edgecolors='black', picker=5,
                       color=elements.check_color(n), label=f"{self.atom[i]}",
                       s=elements.check_radii(n) * 300)

        atoms_pair = []
        for i in range(len(self.bond_list)):
            get_atoms = self.bond_list[i]
            x, y, z = zip(*get_atoms)
            atoms = list(zip(x, y, z))
            atoms_pair.append(atoms)

        # Draw line
        for i in range(len(atoms_pair)):
            merge = list(zip(atoms_pair[i][0], atoms_pair[i][1]))
            x, y, z = merge
            ax.plot(x, y, z, 'k-', color="black", linewidth=2)

        # Set legend
        # Remove duplicate labels in legend.
        # Ref.https://stackoverflow.com/a/26550501/6596684
        handles, labels = ax.get_legend_handles_labels()
        handle_list, label_list = [], []
        for handle, label in zip(handles, labels):
            if label not in label_list:
                handle_list.append(handle)
                label_list.append(label)
        leg = fig.legend(handle_list, label_list,
                         loc="lower left", scatterpoints=1, fontsize=12)
        # Fixed size of point in legend
        # Ref. https://stackoverflow.com/a/24707567/6596684
        for i in range(len(leg.legendHandles)):
            leg.legendHandles[i]._sizes = [90]

        # Set axis
        ax.set_xlabel(r'X', fontsize=15)
        ax.set_ylabel(r'Y', fontsize=15)
        ax.set_zlabel(r'Z', fontsize=15)
        ax.set_title('Full complex', fontsize="12")
        ax.grid(True)

        def insert_text(text, coord, group):
            """
            Insert text in boxes.

            Parameters
            ----------
            text : str
                Text.
            coord : list or array
                Coordinates.
            group : str
                Group A or B.

            """
            if group == "A":
                self.box_1.insert(tk.INSERT, text + "\n")
                self.box_1.see(tk.END)

                self.coord_A.append(coord)

            elif group == "B":
                self.box_2.insert(tk.INSERT, text + "\n")
                self.box_2.see(tk.END)

                self.coord_B.append(coord)

        def on_pick(event):
            """
            Pick point and get XYZ data

            Parameters
            ----------
            event : object
                Event object for on-pick function.

            Examples
            --------
            >>> def on_pick(event):
            >>> ... ind = event.ind
            >>> ... print("on_pick scatter:", ind, np.take(x, ind), np.take(y, ind))

            """
            ind = event.ind[0]
            x, y, z = event.artist._offsets3d
            for i in range(len(self.atom)):
                if x[ind] == self.coord[i][0]:
                    if y[ind] == self.coord[i][1]:
                        if z[ind] == self.coord[i][2]:
                            atom = self.atom[i]
                            break

            results = f"{i + 1}  {atom}  :  {x[ind]} {y[ind]} {z[ind]}"
            coord = [x[ind], y[ind], z[ind]]
            insert_text(results, coord, group)
            # Highlight selected atom
            index = elements.check_atom(atom)
            ax.scatter(x[ind], y[ind], z[ind],
                       marker='o', linewidths=0.5,
                       edgecolors='orange', picker=5, alpha=0.5,
                       color='yellow', s=elements.check_radii(index) * 400)
            # print(i+1, atom, x[ind], y[ind], z[ind])

        fig.canvas.mpl_connect('pick_event', on_pick)

        # plt.axis('equal')
        # plt.axis('off')
        plt.show()

    ########################################
    # Plot fit plant to the selected atoms #
    ########################################

    def plot_fit_plane(self):
        """
        Display complex and two fit planes of two sets of ligand in molecule.

        """
        ###############
        # Clear boxes #
        ###############

        self.box_angle1.delete(0, tk.END)
        self.box_angle2.delete(0, tk.END)

        self.box_eq1.delete(0, tk.END)
        self.box_eq2.delete(0, tk.END)

        ########################
        # Find eq of the plane #
        ########################

        xx, yy, z, abcd = plane.find_fit_plane(self.coord_A)
        plane_a = (xx, yy, z)
        a1, b1, c1, d1 = abcd

        self.box_eq1.insert(tk.INSERT, f"{a1:8.5f}x {b1:+8.5f}y {c1:+8.5f}z {d1:+8.5f} = 0")

        xx, yy, z, abcd = plane.find_fit_plane(self.coord_B)
        plane_b = (xx, yy, z)
        a2, b2, c2, d2 = abcd

        self.box_eq2.insert(tk.INSERT, f"{a2:8.5f}x {b2:+8.5f}y {c2:+8.5f}z {d2:+8.5f} = 0")

        ####################################
        # Calculate angle between 2 planes #
        ####################################

        angle = linear.angle_btw_planes(a1, b1, c1, a2, b2, c2)
        self.box_angle1.insert(tk.INSERT, f"{angle:10.6f}")  # insert to box

        sup_angle = abs(180 - angle)  # supplementary angle
        self.box_angle2.insert(tk.INSERT, f"{sup_angle:10.6f}")  # insert to box

        ###############
        # Plot planes #
        ###############

        fig = plt.figure()
        # fig = plt.figure(figsize=(5, 4), dpi=100)
        ax = Axes3D(fig)
        # ax = fig.add_subplot(111, projection='3d')

        # Plot all atoms
        for i in range(len(self.coord)):
            # Determine atomic number
            n = elements.check_atom(self.atom[i])
            ax.scatter(self.coord[i][0], self.coord[i][1], self.coord[i][2],
                       marker='o', linewidths=0.5, edgecolors='black', picker=5,
                       color=elements.check_color(n), label=f"{self.atom[i]}",
                       s=elements.check_radii(n) * 300)

        atoms_pair = []
        for i in range(len(self.bond_list)):
            get_atoms = self.bond_list[i]
            x, y, z = zip(*get_atoms)
            atoms = list(zip(x, y, z))
            atoms_pair.append(atoms)

        # Draw line
        for i in range(len(atoms_pair)):
            merge = list(zip(atoms_pair[i][0], atoms_pair[i][1]))
            x, y, z = merge
            ax.plot(x, y, z, 'k-', color="black", linewidth=2)

        # Set legend
        # Remove duplicate labels in legend.
        # Ref.https://stackoverflow.com/a/26550501/6596684
        handles, labels = ax.get_legend_handles_labels()
        handle_list, label_list = [], []
        for handle, label in zip(handles, labels):
            if label not in label_list:
                handle_list.append(handle)
                label_list.append(label)
        leg = fig.legend(handle_list, label_list,
                         loc="lower left", scatterpoints=1, fontsize=12)
        # Fixed size of point in legend
        # Ref. https://stackoverflow.com/a/24707567/6596684
        for i in range(len(leg.legendHandles)):
            leg.legendHandles[i]._sizes = [90]

        # Set axis
        ax.set_xlabel(r'X', fontsize=15)
        ax.set_ylabel(r'Y', fontsize=15)
        ax.set_zlabel(r'Z', fontsize=15)
        ax.set_title('Full complex', fontsize="12")
        ax.grid(True)

        # Plot plane A
        xx, yy, z = plane_a
        ax.plot_surface(xx, yy, z, alpha=0.2, color='green')

        # Plot plane B
        xx, yy, z = plane_b
        ax.plot_surface(xx, yy, z, alpha=0.2, color='red')

        # ax.set_xlim(-10, 10)
        # ax.set_ylim(-10, 10)
        # ax.set_zlim(0, 10)

        # plt.axis('equal')
        # plt.axis('off')
        plt.show()

    ##################
    # Clear text box #
    ##################

    def clear_text(self):
        """
        Clear text in box A & B.

        """
        self.box_1.delete(1.0, tk.END)
        self.box_2.delete(1.0, tk.END)

        self.box_angle1.delete(0, tk.END)
        self.box_angle2.delete(0, tk.END)

        self.box_eq1.delete(0, tk.END)
        self.box_eq2.delete(0, tk.END)

        self.coord_A = []
        self.coord_B = []

        try:
            plt.close("all")
        except AttributeError:
            pass

    def show_app(self):
        """
        Show application.

        """
        self.wd.mainloop()


class CalcRMSD:
    """
    Calculate root mean squared displacement of atoms in complex, RMSD.

    Parameters
    ----------
    coord_1 : array_like
        Atomic coordinates of structure 1.
    coord_2 : array_like
        Atomic coordinates of structure 2.
    atom_1 : list, tuple, optional
        Atomic symbols of structure 1.
    atom_2 : list, tuple, optional
        Atomic symbols of structure 2.
        If no atom_2 specified, assign it with None.

    Returns
    -------
    rmsd_normal : float
        Normal RMSD.
    rmsd_translate : float
        Translate RMSD (re-centered).
    rmsd_rotate : float
        Kabsch RMSD (rotated).

    References
    ----------
    .. [2] J. C. Kromann. https://github.com/charnley/rmsd.

    Examples
    --------
    >>> # Example of structure 1
    >>> comp1 = [[10.1873, 5.7463, 5.615],
                 [8.494, 5.9735, 4.8091],
                 [9.6526, 6.4229, 7.3079],
                 [10.8038, 7.5319, 5.1762],
                 [9.6229, 3.9221, 6.0083],
                 [12.0065, 5.5562, 6.3497],
                 [10.8046, 4.9471, 3.9219]]

    >>> # Example of structure 1
    >>> comp2 = [[12.0937, 2.4505, 3.4207],
                 [12.9603, 2.2952, 1.7286],
                 [13.4876, 1.6182, 4.4230],
                 [12.8522, 4.3174, 3.9894],
                 [10.9307, 0.7697, 2.9315],
                 [10.7878, 2.2987, 5.1071],
                 [10.6773, 3.7960, 2.5424]]

    >>> test = CalcRMSD(coord_1=comp1, coord_2=comp2)
    >>> test.calc_rmsd()
    >>> test.rmsd_normal
    6.758144
    >>> test.rmsd_translate
    0.305792
    >>> test.rmsd_rotate
    0.277988

    """

    def __init__(self, coord_1, coord_2, atom_1=None, atom_2=None, master=None, icon=None):
        self.coord_1 = np.asarray(coord_1, dtype=np.float64)
        self.coord_2 = np.asarray(coord_2, dtype=np.float64)

        self.atom_1 = atom_1
        self.atom_2 = atom_2

        if self.atom_2 is None:
            self.atom_2 = self.atom_1

        if master is None:
            self.wd = tk.Tk()
        else:
            self.wd = tk.Toplevel(master)

        self.icon = icon

        self.rmsd_normal = 0
        self.rmsd_translate = 0
        self.rmsd_rotate = 0

    def start_app(self):
        if self.icon is not None:
            self.wd.wm_iconbitmap(self.icon)
        self.wd.title("Calculate RMSD")
        self.wd.geometry("700x440")
        self.wd.resizable(0, 0)

        self.lbl = tk.Label(self.wd, text="Calculate root-mean-square deviation of atomic positions")
        # self.lbl.config(width=12)
        self.lbl.grid(padx="5", pady="5", row=0, column=0, columnspan=2)

        self.lbl = tk.Label(self.wd, text="Structure A")
        self.lbl.config(width=12)
        self.lbl.grid(padx="10", pady="5", row=1, column=0)

        self.lbl = tk.Label(self.wd, text="Structure B")
        self.lbl.config(width=12)
        self.lbl.grid(padx="10", pady="5", row=1, column=1)

        self.box_1 = tkscrolled.ScrolledText(self.wd, height="12", width="45", wrap="word", undo="True")
        self.box_1.grid(padx="5", pady="5", row=2, column=0)

        self.box_2 = tkscrolled.ScrolledText(self.wd, height="12", width="45", wrap="word", undo="True")
        self.box_2.grid(padx="5", pady="5", row=2, column=1)

        self.lbl = tk.Label(self.wd, text="Normal RMSD")
        self.lbl.grid(padx="5", pady="5", row=3, column=0, sticky=tk.E)
        self.box_3 = tk.Entry(self.wd, width="20", justify='center')
        self.box_3.grid(padx="5", row=3, column=1, sticky=tk.W)

        self.lbl = tk.Label(self.wd, text="Re-centered RMSD")
        self.lbl.grid(padx="5", pady="5", row=4, column=0, sticky=tk.E)
        self.box_4 = tk.Entry(self.wd, width="20", justify='center')
        self.box_4.grid(padx="5", row=4, column=1, sticky=tk.W)

        self.lbl = tk.Label(self.wd, text="Rotated RMSD")
        self.lbl.grid(padx="5", pady="5", row=5, column=0, sticky=tk.E)
        self.box_5 = tk.Entry(self.wd, width="20", justify='center')
        self.box_5.grid(padx="5", row=5, column=1, sticky=tk.W)

        self.btn = tk.Button(self.wd, text="Calculate RMSD", command=lambda: self.calc_and_show())
        self.btn.config(width=15, relief=tk.RAISED)
        self.btn.grid(padx="10", pady="15", row=6, column=0, columnspan=2)

        # When app starts, show atomic coordinates in box.
        self.show_coord()

    def show_coord(self):
        """
        Show atomic coordinates in box.

        """
        num_atoms = len(self.coord_1)

        self.box_1.insert(tk.INSERT, f"Total atoms: {num_atoms}\n\n")
        self.box_2.insert(tk.INSERT, f"Total atoms: {num_atoms}\n\n")

        if self.atom_1 is None or self.atom_2 is None:
            self.atom_1 = ['N/A'] * num_atoms
            self.atom_2 = ['N/A'] * num_atoms

        for i in range(num_atoms):
            self.box_1.insert(tk.INSERT, f"{i+1}.  {self.atom_1[i]}\t{self.coord_1[i]}\n")
            self.box_2.insert(tk.INSERT, f"{i+1}.  {self.atom_2[i]}\t{self.coord_2[i]}\n")

    def calc_rmsd(self):
        """
        Calculate normal, translated, and rotated RMSD.

        """
        tmp_1, tmp_2 = self.coord_1.copy(), self.coord_2.copy()

        self.rmsd_normal = rmsd.rmsd(self.coord_1, self.coord_2)

        # Manipulate recenter
        self.coord_1 -= rmsd.centroid(self.coord_1)
        self.coord_2 -= rmsd.centroid(self.coord_2)

        self.rmsd_translate = rmsd.rmsd(self.coord_1, self.coord_2)

        # Rotate
        rotation_matrix = rmsd.kabsch(self.coord_1, self.coord_2)
        self.coord_1 = np.dot(self.coord_1, rotation_matrix)

        self.rmsd_rotate = rmsd.rmsd(self.coord_1, self.coord_2)

        self.coord_1, self.coord_2 = tmp_1, tmp_2

    def calc_and_show(self):
        """
        Execute calc_rmsd function to calculate RMSD
        and show results in box.

        """
        self.calc_rmsd()

        self.box_3.delete(0, tk.END)
        self.box_4.delete(0, tk.END)
        self.box_5.delete(0, tk.END)

        self.box_3.insert(tk.INSERT, f"{self.rmsd_normal: 3.6f}")
        self.box_4.insert(tk.INSERT, f"{self.rmsd_translate: 3.6f}")
        self.box_5.insert(tk.INSERT, f"{self.rmsd_rotate: 3.6f}")

    def show_app(self):
        """
        Show application.

        """
        self.wd.mainloop()

